# Copyright (c) 2019-2021 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from sympy.logic import satisfiable
from sympy.logic import boolalg
import sympy
from typing import Set


class CombinationalOutput:
    """
    Description of an output signal of a combinatorial circuit.
    """

    def __init__(self, function: boolalg.Boolean, high_impedance: boolalg.Boolean):
        self.function: boolalg.Boolean = function
        "Boolean expression for the logic output."
        self.high_impedance: boolalg.Boolean = high_impedance
        "Boolean expression which tells when the output is in high-impedance state."

    def is_tristate(self):
        """
        Check if the output have be high-impedance.
        Check if the high-impedance condition is satisfiable.
        :return: bool
        """
        if self.high_impedance is None:
            return False
        return satisfiable(self.high_impedance)

    def get_inputs(self) -> Set[sympy.Symbol]:
        """
        Find all input pins that are relevant for the outputs.
        :return: Set of input pins.
        """
        pins = set()
        if self.function is not None:
            pins.update(self.function.atoms(sympy.Symbol))
        if self.high_impedance is not None:
            pins.update(self.high_impedance.atoms(sympy.Symbol))
        return pins

    def __str__(self):
        if self.high_impedance:
            return f"CombinationalOutput(f = {self.function}, Z = {self.high_impedance})"
        else:
            return f"CombinationalOutput(f = {self.function})"

    def __repr__(self):
        return str(self)



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Characterization functions for sequential cells.
"""

import os
import tempfile
from itertools import count

from .ngspice_subprocess import simulate_cell

from .util import *
from ..lccommon.net_util import get_subcircuit_ports
from .piece_wise_linear import *

from scipy import optimize
import math

from typing import Dict, List, Optional
import logging

logger = logging.getLogger(__name__)


def ff_find_stabilization_time(
        cell_config: CellConfig,
        cell_ports: List[str],
        clock_input: str,
        data_in: str,
        data_out: str,
        supply_voltage: float,
        setup_time: float,
        clock_edge_polarity: bool,
        rising_data_edge: bool,
        clock_rise_time: float,
        clock_fall_time: float,
        output_load_capacitances: Dict[str, float] = None,
        time_step: float = 100.0e-12,
        max_simulation_time: float = 1e-7,
) -> float:
    """Find the time it takes for the data output signal of a flip-flop to stabilize after an active clock edge.
    This is used to estimate the order of magnitude of the switching speed which will be used in subsequent simulations.
    
    :param cell_ports: All circuit pins/ports in the same ordering as used in the SPICE circuit model.
    :param clock_input: Name of the clock pin ('related pin').
    :param data_in: Name of the data-in pin ('constrained pin').
    :param data_out: Name of the data-out pin.
    :param supply_voltage: Supply voltage in volts.
    :param clock_rise_time: Rise time of the clock signal.
    :param clock_fall_time: Fall time of the clock signal.
    :param output_load_capacitances: A dict with (net, capacitance) pairs which defines the load capacitances attached to certain nets.
    :param time_step: Simulation time step.
    """

    cfg = cell_config.global_conf

    t_clock_edge = time_step * 16  # Rough estimate of when to start the clock edge.

    trip_points = cfg.trip_points

    # Generate the clock edge relative to which the delay will be measured.
    clock_edge = StepWave(
        start_time=t_clock_edge,
        polarity=clock_edge_polarity,
        transition_time=clock_rise_time if clock_edge_polarity else clock_fall_time,
        rise_threshold=trip_points.input_threshold_rise,
        fall_threshold=trip_points.input_threshold_fall
    )
    clock_edge *= supply_voltage

    threshold = 0.5
    breakpoint = f"stop when v({data_out}) > {supply_voltage * threshold} when time > {t_clock_edge}"
    breakpoints = [breakpoint]

    simulation_title = "Estimate flip-flop propagation speed (CLK->D_Out)."

    time, voltages, currents = simulate_cell(
        cell_name=cell_config.cell_name,
        cell_ports=cell_ports,
        input_voltages=input_voltages,
        initial_voltages=initial_conditions,
        breakpoint_statements=breakpoints,
        output_voltages=[data_in, clock_input, data_out],
        output_currents=[cell_config.supply_net],
        simulation_file=sim_file,
        simulation_output_file=sim_output_file,
        max_simulation_time=max_simulation_time,
        simulation_title=simulation_title,
        temperature=cfg.temperature,
        output_load_capacitances=output_load_capacitances,
        time_step=time_step,
        setup_statements=cfg.setup_statements,
        ground_net=cfg.ground_net,
        debug=cfg.debug,
    )

    raise NotImplementedError()


def find_minimum_clock_pulse_width(
        cell_config: CellConfig,
        ff_clock_edge_polarity: bool,
        clock_input: str,
        data_in: str,
        data_out: str,
        setup_time: float,
        clock_pulse_polarity: bool,
        rising_data_edge: bool,
        clock_rise_time: float,
        clock_fall_time: float,
        clock_pulse_width_guess: float,
        output_load_capacitances: Dict[str, float] = None,
        max_delay_estimation: float = 1e-7,
        static_input_voltages: Dict[str, float] = None,
) -> Tuple[float, float]:
    """Find the minimum clock pulse width such that the data is sampled.

    :param ff_clock_edge_polarity: Type of active clock edge. Rising edge: True, Falling edge: False.
    :param rising_data_edge:
    :param clock_pulse_polarity:
    :param setup_time:
    :param cell_config:
    :param clock_pulse_width_guess: Initial value for the minimal clock-pulse length.
        The search for the minimum will be started from there. The closer the guess is, the
        faster the search will terminate.
    :param max_delay_estimation: Upper bound for the clock-to-output delay. For example 10e-9 s.
    :param clock_input: Name of the clock pin ('related pin').
    :param data_in: Name of the data-in pin ('constrained pin').
    :param data_out: Name of the data-out pin.
    :param clock_rise_time: Rise time of the input signal (clock and data).
    :param clock_fall_time: Fall time of the input signal (clock and data).
    :param output_load_capacitances: A dict with (net, capacitance) pairs which defines the load capacitances attached to certain nets.
    :param static_input_voltages: Static input voltages.
        This can be used to set the voltage of static input signals such as scan-enable.
    :return: Returns a tuple of the minimal clock pulse width such that the data signal is sampled and the corresponding delay.
        (min clock pulse width, clock-to-output delay)
    """
    assert isinstance(cell_config, CellConfig)
    cfg = cell_config.global_conf
    assert isinstance(cfg, CharacterizationConfig)
    workingdir = cell_config.workingdir

    logger.debug("Find minimum clock pulse width.")

    supply_voltage = cfg.supply_voltage

    # SPICE include files.
    includes = [f".INCLUDE {cell_config.spice_netlist_file}"]
    includes += cfg.setup_statements

    # Load capacitance statements.
    if output_load_capacitances is None:
        output_load_capacitances = dict()
    else:
        assert isinstance(output_load_capacitances, dict)

    if static_input_voltages is None:
        static_input_voltages = dict()
    else:
        static_input_voltages = static_input_voltages.copy()

    delay_function_cache = dict()

    def delay_function(clock_pulse_width: float, max_allowed_delay: float = max_delay_estimation) -> float:
        """
        Compute the delay from the clock edge to the data output edge.
        If there's no edge at the data output within the maximal simulation time 'Inf' is returned.
        :param clock_pulse_width: Width of the clock pulse.
        :param max_allowed_delay: Maximum allowed clock-to-output delay. The simulation will be stopped
         when the output does not change after `max_allowed_delay` seconds. A delay of `Inf` will be returned.
        :return: Returns the delay from the clock edge to the data output edge or `Inf` if the data edge does not come
        during the maximal allowed delay time.
        """

        def delay_function_uncached(clock_pulse_width: float, max_allowed_delay: float = max_delay_estimation) -> float:

            # Generate the pulse waveform of the clock.
            clock_pulse_shape = PulseWave(
                start_time=0,
                duration=clock_pulse_width,
                polarity=clock_pulse_polarity,
                rise_time=clock_rise_time,
                fall_time=clock_fall_time,
                rise_threshold=cfg.trip_points.input_threshold_rise,
                fall_threshold=cfg.trip_points.input_threshold_fall
            )

            t_pulse1 = setup_time

            clock_pulse1 = clock_pulse_shape.delayed(t_pulse1)

            # Create delayed second pulse.
            time_between_pulses = 4 * pulse_width + 1e-9
            t_pulse2 = t_pulse1 + time_between_pulses
            clock_pulse2 = clock_pulse_shape.delayed(t_pulse2)

            clock_pulse = clock_pulse1 + clock_pulse2
            if not clock_pulse_polarity:
                clock_pulse -= 1

            # Set voltage.
            clock_pulse *= cfg.supply_voltage

            simulation_duration = t_pulse2 + clock_pulse_width + max_allowed_delay

            # Create input signal that first resets the flip-flop to a certain state at the first clock pulse
            # and then switches to the other rail such that the flip-flop changes state with the second clock pulse.

            input_toggle_time = setup_time + time_between_pulses / 2
            input_signal = StepWave(
                start_time=input_toggle_time,
                polarity=rising_data_edge,
                transition_time=1e-15,  # Almost instant.
            )
            input_signal *= cfg.supply_voltage

            # All input voltage signals.
            input_voltages = static_input_voltages.copy()
            input_voltages.update({
                cell_config.supply_net: supply_voltage,
                clock_input: clock_pulse,
                data_in: input_signal
            })

            # Initial voltages of output nodes.
            initial_conditions = {
                data_out: cfg.supply_voltage - input_signal(0)  # The inverse of data_in.
            }

            # Simulate until output signal clearly reaches the other rail.
            # Waiting until the signal reaches ~50% is not sufficient because
            # it might be in a meta stable state.
            threshold = 0.98 if rising_data_edge else 0.02
            cmp = ">" if rising_data_edge else "<"
            breakpoint_statement = f"stop when v({data_out}) {cmp} {supply_voltage * threshold} " \
                                   f"when time > {setup_time + time_between_pulses}"
            breakpoints = [breakpoint_statement]

            # Simulation script file path.
            file_name = f"lctime_min_clk_pulse_width_" \
                        f"{'pos_pulse' if clock_pulse_polarity else 'neg_pulse'}_" \
                        f"{'data_rising' if rising_data_edge else 'data_falling'}"
            sim_file = os.path.join(workingdir, f"{file_name}.sp")

            # Output file for simulation results.
            sim_output_file = os.path.join(workingdir, f"{file_name}_output.txt")
            # File for debug plot of the waveforms.
            sim_plot_file = os.path.join(workingdir, f"{file_name}_plot.svg")

            simulation_title = f"Find minimum clock pulse width: '{data_in}'-'{clock_input}'->'{data_out}', " \
                               f"pulse polarity={clock_pulse_polarity}."

            time, voltages, currents = simulate_cell(
                cell_name=cell_config.cell_name,
                cell_ports=cell_config.spice_ports,
                input_voltages=input_voltages,
                initial_voltages=initial_conditions,
                breakpoint_statements=breakpoints,
                output_voltages=[data_in, clock_input, data_out],
                output_currents=[cell_config.supply_net],
                simulation_file=sim_file,
                simulation_output_file=sim_output_file,
                max_simulation_time=simulation_duration,
                simulation_title=simulation_title,
                temperature=cfg.temperature,
                output_load_capacitances=output_load_capacitances,
                time_step=cfg.time_step,
                setup_statements=includes,
                ground_net=cell_config.ground_net,
                debug=cfg.debug,
            )

            supply_current = currents[cell_config.supply_net]
            input_voltage = voltages[data_in]
            clock_voltage = voltages[clock_input]
            output_voltage = voltages[data_out]

            # Cut away the signals of the setup procedure.
            start_time = input_toggle_time
            assert start_time <= time[-1]
            start_index = np.arange(len(time))[time > start_time][0]
            time = time[start_index:]
            supply_current = supply_current[start_index:]
            input_voltage = input_voltage[start_index:]
            clock_voltage = clock_voltage[start_index:]
            output_voltage = output_voltage[start_index:]

            if cfg.debug_plots:
                # Plot data in debug mode.
                logger.debug("Create plot of waveforms: {}".format(sim_plot_file))
                import matplotlib
                # matplotlib.use('Agg')
                import matplotlib.pyplot as plt
                plt.close()
                plt.title("Clock to output delay")
                plt.plot(time, clock_voltage, label='clock')
                plt.plot(time, input_voltage, label='data_in')
                plt.plot(time, output_voltage, label='data_out')
                plt.plot(time, supply_current, label='supply_current')
                plt.legend()
                plt.savefig(sim_plot_file)
                # plt.show()
                plt.close()

            # Normalize
            clock_voltage /= supply_voltage
            input_voltage /= supply_voltage
            output_voltage /= supply_voltage

            # Turn a falling edge into a rising edge by flipping the signal.
            # This makes measurement of the delay easier.
            if not rising_data_edge:
                output_voltage = 1 - output_voltage

            # Get decision thresholds.
            if rising_data_edge:
                output_threshold = cfg.trip_points.output_threshold_rise
            else:
                output_threshold = cfg.trip_points.output_threshold_fall

            # Get logical values at start and end.
            logic_out_start = output_voltage[0] > output_threshold
            logic_out_end = output_voltage[-1] > output_threshold

            # The delay can only be measured if there is a rising edge in the output.
            if not logic_out_start and logic_out_end:
                # Output has rising edge
                # Get first clock edge.
                thresh_clk = cfg.trip_points.input_threshold_rise if clock_pulse_polarity else cfg.trip_points.input_threshold_fall
                t_active_clock_edge = transition_time(voltage=clock_voltage, time=time, n=0,
                                                      assert_one_crossing=True,
                                                      threshold=thresh_clk,
                                                      find_falling_edges=not ff_clock_edge_polarity,
                                                      find_rising_edges=ff_clock_edge_polarity)

                assert t_active_clock_edge is not None

                # Get first output data edge.
                thresh_data = cfg.trip_points.input_threshold_rise if rising_data_edge else cfg.trip_points.input_threshold_fall
                t_output_data_edge = transition_time(voltage=output_voltage, time=time, n=0,
                                                     threshold=thresh_data)
                assert t_output_data_edge is not None

                # Compute the delay from the clock edge to the output data edge.
                delay = t_output_data_edge - t_active_clock_edge
            else:
                # There's no edge in the output.
                delay = float('Inf')

            return delay

        # Check cache.
        cache_tag = (clock_pulse_width, max_allowed_delay)
        result = delay_function_cache.get(cache_tag)
        if result is None:
            result = delay_function_uncached(clock_pulse_width, max_allowed_delay)
            delay_function_cache[cache_tag] = result

        return result

    pulse_width = clock_pulse_width_guess
    delay_estimation = max_delay_estimation  # Estimation of the order of magnitude of the clock-to-output delay.
    # Find a pulse width that is long enough.
    while True:
        delay = delay_function(pulse_width)
        logger.debug(f"Pulse width = {pulse_width}, Delay = {delay}")
        if math.isinf(delay):
            pulse_width = pulse_width * 2
            assert pulse_width < 1e-6, "Does not converge."
        else:
            assert not math.isinf(delay)
            assert delay > 0
            delay_estimation = delay * 10  # Store the delay as an estimation of the delay.
            break

    # Find a pulse width such that the minimal clock-to-output delay is met.
    delay_prev = delay
    pulse_width = pulse_width * 2
    while True:
        delay = delay_function(pulse_width, delay_estimation)
        err = abs(delay - delay_prev)
        if err > 1e-12:
            pulse_width = pulse_width * 2
            assert pulse_width < 1e-6, "Does not converge."
        else:
            break
        delay_prev = delay

    # Remember the upper bound of the pulse width.
    upper_bound = pulse_width
    reference_delay = delay_function(upper_bound, delay_estimation)
    assert not math.isinf(reference_delay)

    # Find a pulse width that is too short.
    pulse_width = clock_pulse_width_guess
    while True:
        delay = delay_function(pulse_width, delay_estimation)
        logger.debug(f"Pulse width = {pulse_width}, Delay = {delay}")
        if not math.isinf(delay):
            pulse_width = pulse_width / 2
        else:
            break
    lower_bound = pulse_width

    logger.debug(f"Search for minimum pulse width in the interval [{lower_bound}, {upper_bound}].")

    max_delay = reference_delay + cfg.max_pushout_time

    def f(pulse_width) -> float:
        result = delay_function(pulse_width, delay_estimation) - max_delay
        logger.debug(f"f({pulse_width}) == {result}")
        return result

    assert f(lower_bound) > 0, f"f({lower_bound}) = {f(lower_bound)}"
    assert f(upper_bound) < 0, f"f({upper_bound}) = {f(upper_bound)}"

    min_pulse_width = optimize.brentq(f, lower_bound, upper_bound)
    assert isinstance(min_pulse_width, float)

    logger.debug(f"Minimal clock pulse: {min_pulse_width}s")
    corresponding_delay = delay_function(min_pulse_width, delay_estimation)
    assert isinstance(corresponding_delay, float)
    logger.debug(f"Corresponding delay: {corresponding_delay}")
    assert not math.isinf(corresponding_delay)

    return min_pulse_width, corresponding_delay


# def test_find_min_pulse_width():
#     trip_points = TripPoints(
#         input_threshold_rise=0.5,
#         input_threshold_fall=0.5,
#         output_threshold_rise=0.5,
#         output_threshold_fall=0.5,
#
#         slew_lower_threshold_rise=0.2,
#         slew_upper_threshold_rise=0.8,
#         slew_lower_threshold_fall=0.2,
#         slew_upper_threshold_fall=0.8
#     )
#
#     subckt_name = 'DFFPOSX1'
#
#     include_file = f'../../test_data/freepdk45/netlists_pex/{subckt_name}.pex.netlist'
#     model_file = f'../../test_data/freepdk45/gpdk45nm.m'
#
#     ports = get_subcircuit_ports(include_file, subckt_name)
#     print("Ports: ", ports)
#     data_in = 'D'
#     clock = 'CLK'
#     data_out = 'Q'
#     ground = 'GND'
#     supply = 'VDD'
#
#     input_rise_time = 0.000e-9
#     input_fall_time = 0.000e-9
#
#     temperature = 27
#     logger.info(f"Temperature: {temperature} C")
#
#     output_load_capacitances = {data_out: 0.06e-12}
#     logger.info(f"Output load capacitance: {output_load_capacitances} [F]")
#
#     time_step = 10e-12
#     logger.info(f"Time step: {time_step} s")
#
#     # TODO: find appropriate simulation_duration_hint
#     simulation_duration_hint = 250e-12
#
#     # SPICE include files.
#     includes = [include_file, model_file]
#
#     vdd = 1.1
#     logger.info(f"Supply voltage: {vdd} V")
#
#     setup_time = 1e-9  # Choose big enough such that initial disturbances settle down.
#     clock_pulse_polarity = True
#     rising_data_edge = True
#
#     # Voltage sources for input signals.
#     # input_sources = [circuit.V('in_{}'.format(inp), inp, circuit.gnd, 'dc 0 external') for inp in inputs]
#
#     pos_edge_flipflop = True
#
#     def _min_pulse_width(pulse_polarity: bool) -> float:
#         return find_minimum_pulse_width(
#             cell_name=subckt_name,
#             cell_ports=ports,
#             clock_input=clock,
#             data_in=data_in,
#             data_out=data_out,
#             setup_time=setup_time,
#             clock_pulse_polarity=pulse_polarity,
#             rising_data_edge=rising_data_edge,
#             supply_voltage=vdd,
#             input_rise_time=input_rise_time,
#             input_fall_time=input_fall_time,
#             trip_points=trip_points,
#             temperature=temperature,
#             output_load_capacitances=output_load_capacitances,
#             time_step=time_step,
#             setup_statements=includes,
#             ground_net=ground,
#             supply_net=supply,
#             # debug=True
#         )
#
#     clock_pulse_polarity = False
#     min_pulse_width_low = _min_pulse_width(pulse_polarity=clock_pulse_polarity)
#     clock_pulse_polarity = True
#     min_pulse_width_high = _min_pulse_width(pulse_polarity=clock_pulse_polarity)
#
#     print(f"min_pulse_width_high = {min_pulse_width_high}")
#     print(f"min_pulse_width_low = {min_pulse_width_low}")
#     assert isinstance(min_pulse_width_high, float)


def get_clock_to_output_delay(
        cell_conf: CellConfig,
        clock_input: str,
        data_in: str,
        data_out: str,
        setup_time: float,
        hold_time: float,
        rising_clock_edge: bool,
        rising_data_edge: bool,
        input_rise_time: float,
        input_fall_time: float,
        clock_rise_time: float,
        clock_fall_time: float,
        clock_cycle_hint: float,
        enable_pre_measurement_clock_cycle: bool = True,
        output_load_capacitances: Dict[str, float] = None,
        setup_statements: List[str] = None,
        input_voltages: Dict[str, float] = None,
        include_slew: bool = False,
) -> float:
    """Get the delay from the clock edge to the output edge.

    :param input_voltages: Static input voltages.
        This can be used to set the voltage of static input signals such as scan-enable.
    :param clock_input: Name of the clock pin ('related pin').
    :param data_in: Name of the data-in pin ('constrained pin').
    :param data_out: Name of the data-out pin.
    :param setup_time: Delay from data input `D` edge to rising clock edge.
    :param hold_time: Delay from clock edge to data input edge.
    :param rising_clock_edge: `True` = use rising clock edge, `False` = use falling clock edge.
    :param rising_data_edge: `True` = use rising data edge, `False` = use falling data edge.
    :param input_rise_time: Rise time of the input signal (clock and data).
    :param input_fall_time: Fall time of the input signal (clock and data).
    :param output_load_capacitances: A dict with (net, capacitance) pairs which defines the load capacitances attached to certain nets.
    :param clock_cycle_hint: Run the simulation for at least this amount of time.
    :param enable_pre_measurement_clock_cycle: Apply a clock cycle before applying the clock edge.
        Used to initialize the internal state of a flip-flop.
        Set this to `False` for latches.
    :param setup_statements: SPICE statements that are included at the beginning of the simulation.
        This should be used for .INCLUDE and .LIB statements.
    :param include_slew: If set to True return a tuple of (delay time, slew time).
    :return: Returns the delay from the clock edge to the data edge.
     Returns `Inf` if the output does not toggle within the maximum simulation time.
    """

    assert isinstance(cell_conf, CellConfig)
    assert cell_conf.workingdir is not None
    workingdir = cell_conf.workingdir

    cfg = cell_conf.global_conf
    assert isinstance(cfg, CharacterizationConfig)

    trip_points = cfg.trip_points
    assert isinstance(trip_points, TripPoints)

    supply_voltage = cfg.supply_voltage
    supply_net = cell_conf.supply_net
    ground_net = cell_conf.ground_net

    logger.debug("get_clock_to_output_delay() ...")

    # Load include files.
    if setup_statements is None:
        setup_statements = []

    period = max(clock_cycle_hint, input_rise_time + input_fall_time, clock_rise_time + clock_fall_time)

    # Generate the wave form of the clock.
    if enable_pre_measurement_clock_cycle:
        # First a clock pulse makes sure that the right state is sampled into the cell.
        clock_pulse1 = PulseWave(
            start_time=period,
            duration=period,
            polarity=rising_clock_edge,
            rise_time=clock_rise_time,
            fall_time=clock_fall_time,
            rise_threshold=trip_points.input_threshold_rise,
            fall_threshold=trip_points.input_threshold_fall
        )

        t_clock_edge = 4 * period + max(setup_time, 0)
    else:
        clock_pulse1 = 0
        t_clock_edge = period + max(setup_time, 0) # TODO: Choose time as tight as possible

    assert t_clock_edge > 0

    # Generate the clock edge relative to which the delay will be measured.
    clock_transition_time = clock_rise_time if rising_clock_edge else clock_fall_time
    clock_edge = StepWave(
        start_time=t_clock_edge,
        polarity=rising_clock_edge,
        transition_time=clock_transition_time,
        rise_threshold=trip_points.input_threshold_rise,
        fall_threshold=trip_points.input_threshold_fall
    )

    # Sanity check:
    assert math.isclose(clock_edge(t_clock_edge),
                        trip_points.input_threshold_rise if rising_clock_edge
                        else trip_points.input_threshold_fall)

    clk_wave = clock_pulse1 + clock_edge

    if not rising_clock_edge:
        # Compensate for offset.
        clk_wave -= 1

    simulation_end = t_clock_edge + 4 * period

    # Create data pulse.
    logger.debug("Create data pulse.")
    input_wave = PulseWave(
        start_time=t_clock_edge - setup_time,
        duration=setup_time + hold_time,
        polarity=rising_data_edge,
        rise_time=input_rise_time,
        fall_time=input_fall_time,
        rise_threshold=trip_points.input_threshold_rise,
        fall_threshold=trip_points.input_threshold_fall
    )

    input_wave *= supply_voltage
    clk_wave *= supply_voltage

    if input_voltages is None:
        input_voltages = dict()
    else:
        input_voltages = input_voltages.copy()

    # Set the data and clock signals.
    input_voltages.update({
        supply_net: supply_voltage,
        clock_input: clk_wave,
        data_in: input_wave
    })

    # Load capacitance statements.
    if output_load_capacitances is None:
        output_load_capacitances = dict()
    else:
        assert isinstance(output_load_capacitances, dict)

    # Initial voltages of output nodes..
    initial_conditions = {
        data_out: 0 if rising_data_edge else supply_voltage
    }

    # Simulate only until output reaches threshold.
    # Compute stopping voltages of the output signal.
    if rising_data_edge:
        # Rising edge.
        # Add a margin on the threshold to simulate a bit longer.
        threshold = 1 - 0.1 * (1 - trip_points.output_threshold_rise)
    else:
        # Falling edge.
        # Subtract a margin on the threshold to simulate a bit longer.
        threshold = 0.1 * trip_points.output_threshold_fall

    # Set a breakpoint that is active only short before the actual clock edge.
    # This way it ignores signal changes during the initialization.
    cmp = '>' if rising_data_edge else '<'
    assert t_clock_edge - period / 2 > 0
    breakpoint_statement = f"stop when v({data_out}) {cmp} {supply_voltage * threshold} " \
                           f"when time > {t_clock_edge + clock_transition_time * 0.05}"

    breakpoints = [breakpoint_statement]

    # Simulation script file path.
    # file_name = f"lctime_clock_to_output_delay_" \
    #             f"{''.join((f'{net}={v}' for net, v in input_voltages.items()))}_" \
    #             f"{'clk_rising' if rising_clock_edge else 'clk_falling'}_" \
    #             f"{'data_rising' if rising_data_edge else 'data_falling'}"

    file_name = f"lctime_clock_to_output_delay_" \
                f"{'clk_rising' if rising_clock_edge else 'clk_falling'}_" \
                f"{'data_rising' if rising_data_edge else 'data_falling'}"
    sim_file = os.path.join(workingdir, f"{file_name}.sp")

    # Output file for simulation results.
    sim_output_file = os.path.join(workingdir, f"{file_name}_output.txt")
    # File for debug plot of the waveforms.
    sim_plot_file = os.path.join(workingdir, f"{file_name}_plot.svg")

    simulation_title = f"Measure constraint '{data_in}'-'{clock_input}'->'{data_out}', rising_clock_edge={rising_clock_edge}."

    # Choose time step adaptively relative to the smalles input transition time.
    # Upper bound: set by global config.
    time_step = max(cfg.time_step, min(clock_rise_time, input_rise_time) / 128)

    time, voltages, currents = simulate_cell(
        cell_name=cell_conf.cell_name,
        cell_ports=cell_conf.spice_ports,
        input_voltages=input_voltages,
        initial_voltages=initial_conditions,
        breakpoint_statements=breakpoints,
        output_voltages=[data_in, clock_input, data_out],
        output_currents=[supply_net],
        simulation_file=sim_file,
        simulation_output_file=sim_output_file,
        max_simulation_time=simulation_end,
        simulation_title=simulation_title,
        temperature=cfg.temperature,
        output_load_capacitances=output_load_capacitances,
        time_step=time_step,
        setup_statements=setup_statements,
        ground_net=ground_net,
        debug=cfg.debug,
    )

    supply_current = currents[supply_net]
    input_voltage = voltages[data_in]
    clock_voltage = voltages[clock_input]
    output_voltage = voltages[data_out]

    if cfg.debug_plots:
        # Plot data in debug mode.
        logger.debug("Create plot of waveforms: {}".format(sim_plot_file))
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        plt.close()
        plt.title("Clock to output delay")
        plt.plot(time, clock_voltage, label='clock')
        plt.plot(time, input_voltage, label='data_in')
        plt.plot(time, output_voltage, label='data_out')
        plt.plot(time, supply_current, label='supply_current')
        plt.legend()
        plt.savefig(sim_plot_file)
        plt.close()

    # Start of interesting interval
    start_time = t_clock_edge - period / 2
    assert start_time < time[-1]
    start_index = np.arange(len(time))[time > start_time][0]
    # Cut away initialization signals.
    time = time[start_index:]
    clock_voltage = clock_voltage[start_index:]
    input_voltage = input_voltage[start_index:]
    output_voltage = output_voltage[start_index:]

    # import matplotlib.pyplot as plt
    # plt.plot(time, clock_voltage, label='clock')
    # plt.plot(time, input_voltage)
    # plt.plot(time, output_voltage)
    # plt.legend()
    # plt.show()

    # Normalize voltages (divide by VDD).
    clock_voltage /= supply_voltage
    input_voltage /= supply_voltage
    output_voltage /= supply_voltage

    # Turn a falling edge into a rising edge by flipping the signal.
    # This makes measurement of the delay easier.
    if not rising_data_edge:
        output_voltage = 1 - output_voltage

    # Get decision thresholds.
    if rising_data_edge:
        output_threshold = trip_points.output_threshold_rise
    else:
        # Invert threshold because the signal gets also normalized.
        output_threshold = 1 - trip_points.output_threshold_fall

    # Get logical values at start and end.
    logic_out_start = output_voltage[0] > output_threshold
    logic_out_end = output_voltage[-1] > output_threshold

    # The delay can only be measured if there is a rising edge in the output.
    if not logic_out_start and logic_out_end:
        # Output has rising edge
        delay = get_input_to_output_delay(time=time, input_signal=clock_voltage,
                                          output_signal=output_voltage, trip_points=trip_points)
        slew = get_slew_time(time=time, voltage=output_voltage, trip_points=trip_points)
    else:
        # There's no edge in the output. Delay is infinite.
        delay = float('Inf')
        slew = float('Inf')

    if include_slew:
        return delay, slew
    else:
        # Default:
        return delay


def characterize_flip_flop_setup_hold(
        cell_conf: CellConfig,

        data_in_pin: str,
        data_out_pin: str,
        clock_pin: str,
        clock_edge_polarity: bool,

        related_pin_transition: np.ndarray,
        constrained_pin_transition: np.ndarray,

        static_input_voltages: Dict[str, float] = None,
        output_load_capacitance: float = 0,

) -> Dict[str, np.ndarray]:
    """
    Find the setup and hold constraints between the `data_in_pin` and the `clock_pin` such
    that the output on `data_output_pin` is correct.
    This is done for all combinations of output capacitances and input transition times.

    :param cell_conf:
    :param data_in_pin: The constrained pin.
    :param data_out_pin: Data output pin of the flip-flop.
    :param clock_pin: The related pin.
    :param related_pin_transition: List of clock transition times.
    :param constrained_pin_transition: List of input transition times.
    :param static_input_voltages: Static input signals other than VDD/GND.
    This should include clear/preset and scan_enable signals, if there are any.
    :return:
    """

    def f(input_transition_time: float, clock_transition_time: float):
        result = measure_flip_flop_setup_hold(
            cell_conf=cell_conf,
            data_in_pin=data_in_pin,
            data_out_pin=data_out_pin,
            clock_pin=clock_pin,
            clock_edge_polarity=clock_edge_polarity,
            output_load_capacitances={data_out_pin: output_load_capacitance},
            data_rise_time=input_transition_time,
            data_fall_time=input_transition_time,
            clock_transition_time=clock_transition_time,
            static_input_voltages=static_input_voltages
        )

        return (
            # Setup
            result.dependent_setup_time_rise,
            result.dependent_setup_time_fall,
            result.dependent_setup_delay_rise,
            result.dependent_setup_delay_fall,
            # Hold
            result.dependent_hold_time_rise,
            result.dependent_hold_time_fall,
            result.dependent_hold_delay_rise,
            result.dependent_hold_delay_fall,
        )

    f_vec = np.vectorize(f, cache=True)

    xx, yy = np.meshgrid(constrained_pin_transition, related_pin_transition)

    # Evaluate timing on the input_slew*load_capacitance grid.
    (setup_time_rise, setup_time_fall, setup_delay_rise, setup_delay_fall,
     hold_time_rise, hold_time_fall, hold_delay_rise, hold_delay_fall) = f_vec(xx, yy)

    # TODO: Which delay should be chosen? Min, max, average?
    cell_rise = setup_delay_rise
    cell_fall = setup_delay_fall

    return {
        'related_pin_transition': related_pin_transition,
        'constrained_pin_transition': constrained_pin_transition,
        'setup_rise_constraint': setup_time_rise,
        'setup_fall_constraint': setup_time_fall,
        'hold_rise_constraint': hold_time_rise,
        'hold_fall_constraint': hold_time_fall,
        'cell_rise': cell_rise,
        'cell_fall': cell_fall,
    }


class FFSetupHoldResult:
    """
    Bundle setup/hold times and corresponding delays for single-edge triggered flip-flops.
    This values are valid for a certain input slew / output load combination.
    """

    def __init__(self):
        self.dependent_setup_time_rise = 0
        "Setup time for a rising data signal."
        self.dependent_setup_time_fall = 0
        "Setup time for a falling data signal."
        self.dependent_setup_delay_rise = 0
        "Data output delay for a rising data signal when the setup time is `dependent_setup_time_rise`."
        self.dependent_setup_delay_fall = 0
        "Data output delay for a falling data signal when the setup time is `dependent_setup_time_fall`."

        self.dependent_hold_time_rise = 0
        "Hold time for a rising data signal."
        self.dependent_hold_time_fall = 0
        "Hold time for a falling data signal."
        self.dependent_hold_delay_rise = 0
        "Data output delay for a rising data signal when the hold time is `dependent_hold_time_rise`."
        self.dependent_hold_delay_fall = 0
        "Data output delay for a falling data signal when the hold time is `dependent_hold_time_fall`."

        self.minimum_delay_rise = 0
        """
        Absolute minimal clock-to-output delay that can be achieved with large setup and hold windows.
        """
        self.minimum_delay_fall = 0
        """
        Absolute minimal clock-to-output delay that can be achieved with large setup and hold windows.
        """


def measure_flip_flop_setup_hold(
        cell_conf: CellConfig,
        data_in_pin: str,
        data_out_pin: str,
        clock_pin: str,
        clock_edge_polarity: bool,
        output_load_capacitances: Dict[str, float],
        data_rise_time: float,
        data_fall_time: float,
        clock_transition_time: float,
        hold_margin: float = 10e-12,
        setup_margin: float = 10e-12,
        static_input_voltages: Dict[str, float] = None
) -> FFSetupHoldResult:
    """
    Measure constraints (setup, hold) of a flip-flop.
    :param cell_conf:
    :param hold_margin: Safety margin to put on the hold times. In seconds. Default is 1e-12s.
        Should be larger than 0. Possibly in the order of a pico-second.
    :param setup_margin: Safety margin to put on the hold times. In seconds. Default is 1e-12s.
        Should be larger than 0. Possibly in the order of a pico-second.
    :return: Returns a `FFSetupHoldResult` object that bundles the results.
    """
    assert isinstance(cell_conf, CellConfig)
    cfg = cell_conf.global_conf

    assert hold_margin > 0
    assert setup_margin > 0

    # TODO: find appropriate simulation_duration_hint
    clock_cycle_hint = 1e-9
    logger.debug(f"simulation_duration_hint = {clock_cycle_hint}")

    clock_rise_time = clock_transition_time
    clock_fall_time = clock_transition_time

    logger.debug(f"Output load capacitance: {output_load_capacitances} [F]")

    # SPICE include files.
    includes = [f".INCLUDE {cell_conf.spice_netlist_file}"]
    includes += cfg.setup_statements

    vdd = cfg.supply_voltage

    # Cache for faster re-evaluation of `delay_f`
    cache = dict()

    def delay_f(
            setup_time: float,
            hold_time: float,
            rising_clock_edge: bool,
            rising_data_edge: bool
    ):
        """
        Wrapper around `get_clock_to_output_delay()`. Results are cached such that a further call with same arguments returns the
        cached value of the first call.
        :param setup_time:
        :param hold_time:
        :param rising_clock_edge:
        :param rising_data_edge:
        :return:
        """
        logger.debug(f"evaluate delay_f({setup_time}, {hold_time}, {rising_clock_edge}, {rising_data_edge})")

        cache_tag = (setup_time, hold_time, rising_clock_edge, rising_data_edge)
        result = cache.get(cache_tag)
        if result is None:
            result = get_clock_to_output_delay(
                cell_conf=cell_conf,
                clock_input=clock_pin,
                data_in=data_in_pin,
                data_out=data_out_pin,
                setup_time=setup_time,
                hold_time=hold_time,
                rising_clock_edge=rising_clock_edge,
                rising_data_edge=rising_data_edge,
                input_rise_time=data_rise_time,
                input_fall_time=data_fall_time,
                clock_rise_time=clock_rise_time,
                clock_fall_time=clock_fall_time,
                output_load_capacitances=output_load_capacitances,
                clock_cycle_hint=clock_cycle_hint,
                setup_statements=includes,
                input_voltages=static_input_voltages,
            )
            cache[cache_tag] = result
        else:
            logger.debug('Cache hit.')
        return result

    def find_min_data_delay(rising_data_edge: bool, abstol: float = 1e-12) -> Tuple[float, Tuple[float, float]]:
        """ Find minimum clock->data delay (with large setup/hold window).

        Procedure is as follows: Setup and hold time are increased until the data delay reaches a stable value.

        The tolerance should be chosen in the order of the accepted absolute error.
        """

        # TODO: Find a estimate start value for setup and hold times.
        setup_time_guess = data_rise_time
        hold_time_guess = data_fall_time

        setup_time = setup_time_guess
        hold_time = hold_time_guess

        assert setup_time != 0, "Setup time guess cannot be 0."  # Does not terminate otherwise.
        assert hold_time != 0, "Hold time guess cannot be 0."  # Does not terminate otherwise.

        prev_delay = None
        delay = None
        ctr = count()
        for _ in ctr:
            delay = delay_f(setup_time, hold_time,
                            rising_clock_edge=clock_edge_polarity,
                            rising_data_edge=rising_data_edge)

            if prev_delay is not None and delay != float('Inf'):
                diff = abs(delay - prev_delay)
                if diff < abstol:
                    # Close enough.
                    break
            setup_time = setup_time * 2
            hold_time = hold_time * 2

            assert setup_time < 1e-6, "Does not converge."
            assert hold_time < 1e-6, "Does not converge."

            prev_delay = delay

        logger.debug(f"Minimum clock to data delay: {delay}. "
                    f"(Iterations = {next(ctr)}, "
                    f"setup_time = {setup_time}, hold_time = {hold_time}, "
                    f"absolute error = {diff})")

        # Return the minimum delay and setup/hold times that lead to it.
        # setup/hold times are divided by 2 because the previous values actually lead to a delay that is close enough.
        return delay, (setup_time / 2, hold_time / 2)

    # Find the minimum data delays.
    # They are used to determine the target data delays when finding
    # setup and hold times.
    logger.debug("Find minimal propagation delays.")
    tolerance = 1e-12
    min_rise_delay, (setup_guess_rise, hold_guess_rise) = find_min_data_delay(rising_data_edge=True, abstol=tolerance)
    min_fall_delay, (setup_guess_fall, hold_guess_fall) = find_min_data_delay(rising_data_edge=False, abstol=tolerance)

    logger.debug(f"min_rise_delay = {min_rise_delay}")
    logger.debug(f"min_fall_delay = {min_fall_delay}")

    logger.debug(f"setup_guess_rise = {setup_guess_rise}")
    logger.debug(f"hold_guess_rise = {hold_guess_rise}")

    logger.debug(f"max. allowed clock-to-output push-out time = {cfg.max_pushout_time}")

    # Define flip flop failure: FF fails if delay is larger than max_accepted_{rise,fall}_delay
    max_rise_delay = min_rise_delay + cfg.max_pushout_time
    max_fall_delay = min_fall_delay + cfg.max_pushout_time

    # Compute the closest delay from rising to falling data edge based on the slew.
    dr = cfg.trip_points.slew_upper_threshold_rise - cfg.trip_points.slew_lower_threshold_rise
    edge_duration_rise = data_rise_time / dr  # Time of edge going fully from 0 to 1.
    df = cfg.trip_points.slew_upper_threshold_fall - cfg.trip_points.slew_lower_threshold_fall
    edge_duration_fall = data_fall_time / df  # Time of edge going fully from 1 to 0.

    min_data_edge_separation = edge_duration_rise * (1 - cfg.trip_points.input_threshold_rise) + \
                               edge_duration_fall * (1 - cfg.trip_points.input_threshold_fall)
    # min_data_edge_separation *= 1.01
    # print(f"dr = {dr}")
    # print(f"df = {df}")
    # print(f"edge_duration_rise = {edge_duration_rise}")
    # print(f"min_separation = {min_data_edge_separation}")

    # Tolerances for bisection root finding algorithm.
    xtol = 1e-15
    rtol = 1e-6

    def find_min_setup(rising_data_edge: bool,
                       hold_time: float) -> Tuple[float, float]:
        """
        Find minimal setup time given a fixed hold time.
        Set `hold_time` to a very large value to find the independent minimal setup time.
        :param rising_data_edge: True = rising data edge, False = falling data edge.
        :param hold_time: Fixed hold time.
        :return:
        """
        max_delay = max_rise_delay if rising_data_edge else max_fall_delay
        setup_guess = setup_guess_rise if rising_data_edge else setup_guess_fall

        logger.debug(f"Find min. setup time. Hold time = {hold_time}. rising_data_edge = {rising_data_edge}")

        def f(setup_time: float) -> float:
            """
            Optimization function objective.
            For finding `setup_time` such that the delay equals the maximum allowed delay.
            :param setup_time:
            :return:
            """
            # print('eval f', setup_time)
            # assert setup_time + hold_time >= input_rise_time + input_fall_time
            delay = delay_f(setup_time=setup_time, hold_time=hold_time,
                            rising_clock_edge=clock_edge_polarity,
                            rising_data_edge=rising_data_edge)
            return delay - max_delay

        # Determine min and max setup time for binary search.
        # shortest = -hold_time + data_rise_time + data_fall_time
        shortest = -hold_time
        longest = max(setup_guess_rise, shortest)
        assert shortest <= longest
        a = f(shortest)
        b = f(longest)
        # Make sure that sign(f(a)) != sign(f(b)) such that the zero can be found with a binary search.
        while not a > 0:
            shortest = shortest / 2
            a = f(shortest)

        assert a > 0

        while not b < 0:
            longest = longest * 2
            assert longest < 1e-6, "Does not converge."
            b = f(longest)

        assert b < 0

        # if cfg.debug_plots:
        # Plot data in debug mode.
        # logger.debug("Create plot of waveforms: {}".format(sim_plot_file))
        # import matplotlib
        # matplotlib.use('Agg')
        # import matplotlib.pyplot as plt
        #
        # t_su = np.linspace(shortest, longest, num=100)
        # err = np.vectorize(f)(t_su)
        # plt.plot(t_su, err)
        # plt.show()

        min_setup_time = optimize.brentq(f, shortest, longest, xtol=xtol, rtol=rtol)
        assert isinstance(min_setup_time, float)
        if math.isclose(min_setup_time, shortest) or math.isclose(min_setup_time, longest):
            logger.warning("Result of binary search is on bounds. Optimal setup-time not found.")

        delay_err = f(min_setup_time)
        # Check if we really found the root of `f`.
        logger.debug(f"min_setup_time = {min_setup_time}, delay_err = {delay_err}, max_delay = {max_delay}")
        assert np.allclose(0, delay_err, atol=10e-12), "Failed to find solution for minimal setup time." \
                                                       " Try to decrease the simulation time step."

        return min_setup_time, delay_err + max_delay

    def find_min_hold(rising_data_edge: bool,
                      setup_time: float) -> Tuple[float, float]:
        """
        Find minimal hold time given a fixed setup time.
        Set `setup_time` to a very large value to find the independent minimal hold time.
        :param rising_data_edge: True = rising data edge, False = falling data edge.
        :param setup_time: Fixed setup time.
        :return: Minimal hold time.
        """
        max_delay = max_rise_delay if rising_data_edge else max_fall_delay
        hold_guess = hold_guess_rise if rising_data_edge else hold_guess_fall

        logger.debug(f"Find min. hold time. Setup time = {setup_time}. rising_data_edge = {rising_data_edge}")

        def f(hold_time: float) -> float:
            """
            Function to find zero.
            :param hold_time:
            :return:
            """
            # print('eval f', hold_time)
            delay = delay_f(setup_time=setup_time,
                            hold_time=hold_time,
                            rising_clock_edge=clock_edge_polarity,
                            rising_data_edge=rising_data_edge)
            return delay - max_delay

        # Determine min and max hold time for binary search.
        # shortest = -setup_time + data_rise_time + data_fall_time
        shortest = -setup_time
        longest = hold_guess
        a = f(shortest)
        b = f(longest)
        assert a > 0
        # Make sure that sign(f(a)) != sign(f(b)) such that the zero can be found with a binary search.
        while not b < 0:
            longest = longest * 2
            assert longest < 1e-6, "Does not converge."
            b = f(longest)

        assert b < 0

        min_hold_time = optimize.brentq(f, shortest, longest, xtol=xtol, rtol=rtol)
        assert isinstance(min_hold_time, float)
        delay_err = f(min_hold_time)
        # Check if we really found the root of `f`.
        logger.debug(f"min_hold_time = {min_hold_time}, delay_err = {delay_err}, max_delay = {max_delay}")
        assert np.allclose(0, delay_err, atol=10e-12), "Failed to find solution for minimal hold time." \
                                                       " Try to decrease the simulation time step."

        return min_hold_time, f(min_hold_time) + max_delay

    def find_min_setup_plus_hold(rising_data_edge: bool, setup_guess: float, hold_guess: float) -> Tuple[float, float]:
        """
        Find the smallest `setup_time + hold_time` such that the flip-flop is operational and the output delay
        corresponds to the maximum allowed delay.

        Proceeds as follows:
        1) Start with a guess of the time window. Set window center to 0.
        2) Shift the center of the window such that the output delay is minimized. Update the window center.
        3) Find the window width such that the resulting delay equals the maximum allowed delay. Update the window width.
        4) If tolerance is not met: Continue at 2. Otherwise return the result.

        :param rising_data_edge: True = rising data edge, False = falling data edge.
        :param setup_guess: Initial value for the setup time.
        :param hold_guess: Initial value for the hold time.
        :return:
        """
        max_delay = max_rise_delay if rising_data_edge else max_fall_delay

        logger.debug(f"Find min. setup plus hold time."
                    f"Initial setup = {setup_guess}, initial hold = {hold_guess}. "
                    f"rising_data_edge = {rising_data_edge}")

        def f(window_width: float, center: float) -> float:
            """
            Optimization function objective.
            Setup and hold are represented as a time window around the clock edge where the data signal needs
            to be stable.
            The clock edge is assumed to come at time 0. The shift relative to the clock edge
            is denoted by `center`.
            :param window_width: setup_time + hold_time
            :param center: (hold_time - setup_time) / 2
            :return:
            """
            setup_time = window_width / 2 - center
            hold_time = center + window_width / 2
            delay = delay_f(setup_time=setup_time, hold_time=hold_time,
                            rising_clock_edge=clock_edge_polarity,
                            rising_data_edge=rising_data_edge)
            return delay - max_delay

        center = 0
        window_width = setup_guess + hold_guess

        max_iter = 100  # Limit the number of iterations.
        for i in range(100):
            logger.debug(f"Find minimal setup+hold: iteration {i}")
            # Remember window width to find out when to terminate the loop.
            previous_width = window_width

            # Minimize the output delay by adjusting the center of the window.
            def f_center(center: float) -> float:
                return f(window_width, center)

            center_opt_result = optimize.minimize_scalar(
                f_center,
                bounds=[center - window_width / 2, center + window_width / 2],
                method='bounded',
                tol=xtol
            )
            assert center_opt_result.success, f"Optimization failed: {center_opt_result.message}"
            center = center_opt_result.x

            def f_width(window_width: float):
                return f(window_width, center)

            # Find window width such that the output delay is equal to the maximum allowed output delay.

            width_lower_bound = window_width / 2
            width_upper_bound = window_width

            a = f_width(width_lower_bound)
            b = f_width(width_upper_bound)

            # Find the two bounds for the bisection search.
            # Make sure that sign(f(a)) != sign(f(b)) such that the zero can be found with a binary search.
            while a <= 0:
                width_lower_bound *= 0.5
                a = f_width(width_lower_bound)

            while b >= 0:
                width_upper_bound *= 2
                b = f_width(width_upper_bound)

            assert b < 0 and a > 0

            new_window_width = optimize.brentq(f_width, width_lower_bound, width_upper_bound, xtol=xtol, rtol=rtol)
            assert isinstance(new_window_width, float)
            window_width = new_window_width

            step = window_width - previous_width

            if abs(step) < 1e-14:  # TODO: What tolerance to choose here?
                setup_time = window_width / 2 - center
                hold_time = center + window_width / 2
                logger.debug(f"Found minimal setup + hold: "
                             f"setup = {setup_time}, "
                             f"hold = {hold_time}, "
                             f"setup+hold={window_width}")
                return setup_time, hold_time

        assert False, f"Iteration limit reached: {max_iter}"

    if False:
        logger.info("Find minimal setup + hold time.")
        min_setup_plus_hold = find_min_setup_plus_hold(rising_data_edge=True,
                                                       setup_guess=setup_guess_rise,
                                                       hold_guess=hold_guess_rise
                                                       )
        min_setup, min_hold = min_setup_plus_hold
        logger.info(f"Minimal setup+hold: setup = {min_setup}, hold = {min_hold}")

        # TODO: Store minimal setup+hold

    logger.info("Measure unconditional minimal setup time.")

    """
    The unconditional minimal setup time is the minimal setup time that can be achieved
    when the hold time is infinitely long, i.e. when the input data signal remains
    stable forever after the clock edge.
    """
    hold_time_guess = max(hold_guess_rise, hold_guess_fall) * 8
    min_setup_time_uncond_rise, min_setup_delay_rise = find_min_setup(rising_data_edge=True,
                                                                      hold_time=hold_time_guess)
    min_setup_time_uncond_fall, min_setup_delay_fall = find_min_setup(rising_data_edge=False,
                                                                      hold_time=hold_time_guess)

    logger.info(f"unconditional min. setup time rise: {min_setup_time_uncond_rise}")
    logger.info(f"unconditional min. setup time fall: {min_setup_time_uncond_fall}")
    logger.info(f"max delays (rise): {min_setup_delay_rise}")
    logger.info(f"max delays (fall): {min_setup_delay_fall}")

    logger.info("Measure unconditional minimal hold time.")
    """
    The unconditional minimal hold time is the minimal hold time that can be achieved
    when the setup time is infinitely long, i.e. when the input data signal is already stable
    since an infinite time before the clock edge.
    """
    setup_time_guess = max(setup_guess_rise, setup_guess_fall) * 8  # TODO: How to choose the initial value?
    min_hold_time_uncond_rise, min_hold_delay_rise = find_min_hold(rising_data_edge=True,
                                                                   setup_time=setup_time_guess)
    min_hold_time_uncond_fall, min_hold_delay_fall = find_min_hold(rising_data_edge=False,
                                                                   setup_time=setup_time_guess)

    logger.info(f"unconditional min. hold time rise: {min_hold_time_uncond_rise}")
    logger.info(f"unconditional min. hold time fall: {min_hold_time_uncond_fall}")
    logger.info(f"max delays (rise): {min_hold_delay_rise}")
    logger.info(f"max delays (fall): {min_hold_delay_fall}")

    # Find dependent setup time.
    logger.info("Measure dependent minimal setup time.")
    dependent_setup_time_rise, dependent_setup_delay_rise = \
        find_min_setup(rising_data_edge=True,
                       hold_time=min_hold_time_uncond_rise + hold_margin)

    dependent_setup_time_fall, dependent_setup_delay_fall = \
        find_min_setup(rising_data_edge=False,
                       hold_time=min_hold_time_uncond_fall + hold_margin)

    # Find dependent hold time.
    logger.info("Measure dependent minimal hold time.")
    dependent_hold_time_rise, dependent_hold_delay_rise = \
        find_min_hold(rising_data_edge=True,
                      setup_time=min_setup_time_uncond_rise + setup_margin)

    dependent_hold_time_fall, dependent_hold_delay_fall = \
        find_min_hold(rising_data_edge=False,
                      setup_time=min_setup_time_uncond_fall + setup_margin)

    logger.info(f"dep. setup (rise, fall): {dependent_setup_time_rise}, {dependent_setup_time_fall}")
    logger.info(f"dep. setup delay (rise, fall): {dependent_setup_delay_rise}, {dependent_setup_delay_fall}")

    logger.info(f"dep. hold (rise, fall): {dependent_hold_time_rise}, {dependent_hold_time_fall}")
    logger.info(f"dep. hold delay (rise, fall): {dependent_hold_delay_rise}, {dependent_hold_delay_fall}")

    result = FFSetupHoldResult()

    result.dependent_setup_time_rise = dependent_setup_time_rise
    result.dependent_setup_time_fall = dependent_setup_time_fall
    result.dependent_setup_delay_rise = dependent_setup_delay_rise
    result.dependent_setup_delay_fall = dependent_setup_delay_fall

    result.dependent_hold_time_rise = dependent_hold_time_rise
    result.dependent_hold_time_fall = dependent_hold_time_fall
    result.dependent_hold_delay_rise = dependent_hold_delay_rise
    result.dependent_hold_delay_fall = dependent_hold_delay_fall

    result.minimum_delay_rise = min_rise_delay
    result.minimum_delay_fall = min_fall_delay

    return result

def measure_clock_to_output_delay(
        cell_conf: CellConfig,
        data_in_pin: str,
        data_out_pin: str,
        clock_pin: str,
        clock_edge_polarity: bool,
        data_edge_polarity: bool,
        output_load_capacitances: Dict[str, float],
        clock_transition_time: float,
        data_transition_time: float,
        setup_time: float,
        hold_time: float,
        clock_cycle_hint: float = 1e-9,
        static_input_voltages: Dict[str, float] = None
) -> (float, float):
    """
    Measure the delay and slew from the clock edge to the output data edge.
    
    Returns (delay, slew time)
    """
    assert isinstance(cell_conf, CellConfig)
    cfg = cell_conf.global_conf

    assert setup_time + hold_time >= 0, "sum of setup and hold time must be positive"
    assert data_in_pin != data_out_pin, f"data input ({data_in_pin}) and output pin ({data_out_pin}) must be different"
    
    # TODO: find appropriate simulation_duration_hint
    logger.debug(f"simulation_duration_hint = {clock_cycle_hint}")

    clock_rise_time = clock_transition_time
    clock_fall_time = clock_transition_time

    data_rise_time = data_transition_time
    data_fall_time = data_transition_time

    logger.debug(f"Output load capacitance: {output_load_capacitances} [F]")

    # SPICE include files.
    includes = [f".INCLUDE {cell_conf.spice_netlist_file}"]
    includes += cfg.setup_statements

    vdd = cfg.supply_voltage
    
    return get_clock_to_output_delay(
        cell_conf=cell_conf,
        clock_input=clock_pin,
        data_in=data_in_pin,
        data_out=data_out_pin,
        setup_time=setup_time,
        hold_time=hold_time,
        rising_clock_edge=clock_edge_polarity,
        rising_data_edge=data_edge_polarity,
        input_rise_time=data_rise_time,
        input_fall_time=data_fall_time,
        clock_rise_time=clock_rise_time,
        clock_fall_time=clock_fall_time,
        output_load_capacitances=output_load_capacitances,
        clock_cycle_hint=clock_cycle_hint,
        setup_statements=includes,
        input_voltages=static_input_voltages,
        include_slew=True,
    )



# Copyright (c) 2019 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import itertools
import networkx as nx
from typing import Any, Dict, List, Iterable, Tuple, Set
import collections
import sympy
from sympy.logic import satisfiable, simplify_logic as sympy_simplify_logic
from sympy.logic import boolalg

from ..lccommon.data_types import ChannelType
import logging

from .types import CombinationalOutput

# logging.basicConfig(level=logging.INFO, stream=sys.stdout)
logger = logging.getLogger(__name__)

"""
Extract logic formulas and memory loops from a transistor-level circuit graph.
"""


class Memory:
    """
    Data structure for a memory loop (latch).
    """

    def __init__(self,
                 data: boolalg.Boolean,
                 write_condition: boolalg.Boolean,
                 oscillation_condition: boolalg.Boolean):
        """

        :param data: The input data signal.
        :param write_condition: The condition that makes this loop transparent for new input data.
        When the write condition is `False` then the loop stores the input data.
        :param oscillation_condition: When the oscillation condition is met the loop oscillates or causes a short circuit.
        """
        self.data = data
        self.write_condition = write_condition
        self.oscillation_condition = oscillation_condition
        # ID of the memory loop.
        self.loop_id = None

    def __str__(self):
        return f"Memory(data = {self.data}, write = {self.write_condition}, " \
               f"loop_id = {self.loop_id}, oscillate = {self.oscillation_condition})"

    def __repr__(self):
        return str(self)


class AbstractCircuit:
    """
    Describe the circuit as a set of boolean formulas and latches.
    """

    def __init__(self,
                 output_pins: Set[sympy.Symbol],
                 output_formulas: Dict[boolalg.BooleanAtom, CombinationalOutput],
                 latches: Dict[boolalg.BooleanAtom, Memory]):
        """
        Describe the circuit as a set of boolean formulas and latches.
        Holds the result of `analyze_circuit_graph`.

        :param output_pins: Set of all output pins.
        :param output_formulas: Dictionary that maps output symbols to their boolean formulas.
        :param latches: Dictionary that holds the memory objects for each signal that is driven by a memory/latch.
        """
        self.output_pins: Set[sympy.Symbol] = output_pins
        "Set of all output pins."
        self.outputs: Dict[sympy.Symbol, CombinationalOutput] = output_formulas
        "Dictionary that maps output symbols to their boolean formulas."
        self.latches: Dict[boolalg.BooleanAtom, Memory] = latches
        "Dictionary that holds the memory objects for each signal that is driven by a memory/latch."

    def is_sequential(self) -> bool:
        """
        Test if there are latches in this circuit.
        :return:
        """
        return len(self.latches) > 0

    def get_primary_inputs(self) -> Set[sympy.Symbol]:
        """
        Find all input pins that are relevant for the outputs.
        :return: Set of input pins.
        """
        pins = set()
        for o, f in self.outputs.items():
            pins.update(f.function.atoms(sympy.Symbol))
            pins.update(f.high_impedance.atoms(sympy.Symbol))
        for _, latch in self.latches.items():
            pins.update(latch.write_condition.atoms(sympy.Symbol()))
            pins.update(latch.data.atoms(sympy.Symbol()))
            pins.update(latch.oscillation_condition.atoms(sympy.Symbol()))
        return pins


def simplify_logic(f: boolalg.Boolean, force: bool = True) -> boolalg.Boolean:
    """
    Simplify a boolean formula.
    :param f:
    :param force:
    :return:
    """
    logger.debug(f"Simplify formula: {f}")
    return sympy_simplify_logic(f, force=force)


def simplify_with_assumption(assumption: boolalg.Boolean, formula: boolalg.Boolean) -> boolalg.Boolean:
    """
    Simplify a formula while assuming that `assumption` is satisfied.
    The simplification might be incorrect when the assumption is not satisfied.
    In other words all variable assignments that invalidate the assumption are treated as don't-cares.
    :param assumption:
    :param formula: The formula to be simplified.
    :return: Simplified formula under the given assumption.
    """

    # Extract variables.
    all_vars = set(assumption.atoms(sympy.Symbol)) | set(formula.atoms(sympy.Symbol))

    all_vars = sorted(all_vars, key=lambda a: a.name)

    # Create a truth table.
    # Treat all rows that don't satisfy the assumption as "don't care".
    truth_table_assumption = boolalg.truth_table(assumption, all_vars)
    truth_table_formula = boolalg.truth_table(formula, all_vars, input=False)

    minterms = []
    dont_cares = []
    for (inputs, value_assumption), value_formula in zip(truth_table_assumption, truth_table_formula):

        if not value_assumption:
            dont_cares.append(inputs)
        elif value_formula:
            minterms.append(inputs)

    f = boolalg.SOPform(variables=all_vars, minterms=minterms, dontcares=dont_cares)

    return f


def bool_equals(f1: boolalg.Boolean, f2: boolalg.Boolean) -> bool:
    """
    Check equality of two boolean formulas.
    :param f1:
    :param f2:
    :return:
    """
    return not satisfiable(f1 ^ f2)


def boolean_derivatives(f: boolalg.Boolean, x: sympy.Symbol) \
        -> Tuple[boolalg.Boolean, boolalg.Boolean, boolalg.Boolean]:
    """
    Calculate the boolean derivative, positive derivative and negative derivative.
    :param f:
    :param x:
    :return: (derivative, positive derivative, negative derivative)
    """
    # TODO: Use this also for is_unate_in_xi in util.
    assert isinstance(x, sympy.Symbol)

    f0 = simplify_logic(f.subs({x: False}))
    f1 = simplify_logic(f.subs({x: True}))
    positive_derivative = simplify_logic(~f0 & f1)
    negative_derivative = simplify_logic(f0 & ~f1)

    derivative = simplify_logic(positive_derivative | negative_derivative)  # == f0 ^ f1
    logger.debug('calculate derivative: d/d{} ({})) = {}'.format(x, f, derivative))
    logger.debug('calculate pos. derivative: d+/d{} ({})) = {}'.format(x, f, positive_derivative))
    logger.debug('calculate neg. derivative: d-/d{} ({})) = {}'.format(x, f, negative_derivative))
    return derivative, positive_derivative, negative_derivative


def _get_gate_nets(graph: nx.MultiGraph) -> Set:
    """
    Return a set of all net names that connect to a transistor gate.
    Gate net names are stored as keys of the edges in the multi-graph.
    :param graph:
    :return: Set of net names.
    """
    all_gate_nets = {net_name for (_a, _b, (net_name, _channel_type)) in graph.edges(keys=True)}
    return all_gate_nets


def find_input_gates(graph: nx.MultiGraph) -> Set:
    """
    Find names of input signals.
    Every net that is connected only to transistor gates is considered an input to the cell.
    :param graph:
    :return: Set of input signal names.
    """

    all_gate_nets = _get_gate_nets(graph)
    all_nodes = set(graph.nodes)
    input_nets = all_gate_nets - all_nodes

    return input_nets


def test_find_input_gates():
    g = nx.MultiGraph()
    g.add_edge('vdd', 'nand', ('a', ChannelType.PMOS))
    g.add_edge('vdd', 'nand', ('b', ChannelType.PMOS))
    g.add_edge('gnd', '1', ('a', ChannelType.NMOS))
    g.add_edge('1', 'nand', ('b', ChannelType.NMOS))
    g.add_edge('vdd', 'output', ('nand', ChannelType.PMOS))
    g.add_edge('gnd', 'output', ('nand', ChannelType.NMOS))

    inputs = find_input_gates(g)
    assert inputs == {'a', 'b'}


def _all_simple_paths_multigraph(graph: nx.MultiGraph, source, target, cutoff=None):
    """
    Enumerate all simple paths (no node occurs more than once) from source to target.
    Yields edges inclusive keys of all simple paths in a multi graph.
    :param graph: 
    :param source: Source node.
    :param target: Target node.
    :param cutoff: Stopping condition. Maximal length of the path.
    :return: Generator object.
    """

    # Sanity check: source and target node must be in the graph.
    if source not in graph:
        raise nx.NodeNotFound('source node %s not in graph' % source)
    if target not in graph:
        raise nx.NodeNotFound('target node %s not in graph' % target)

    # Trivial solution.
    if source == target:
        return []

    # Set the cutoff to the number of nodes in the graph minus 1.
    if cutoff is None:
        cutoff = len(graph) - 1

    # If the cutoff is 0, return an empty list.
    if cutoff < 1:
        return []

    # Set of nodes that have been visited already.
    # An ordered dict is used such that nodes can be removed from the set
    # according to the reversed insertion order.
    # This works basically as a stack but allows for faster membership test than a list.
    visited = collections.OrderedDict.fromkeys([source])

    edges = list()  # Store sequence of edges.

    # Put on the stack an iterator over all edges adjacent to the source node.
    stack = [
        (((u, v), key) for u, v, key in graph.edges(source, keys=True))
    ]
    while stack:
        children = stack[-1]  # Take the top of the stack.
        # Take the next element of the iterator.
        (child_source, child), child_key = next(children, ((None, None), None))
        if child is None:
            # The iterator on top of the stack is empty.
            stack.pop()  # Drop the empty iterator.
            # Rewind to the previous 'recursion' level.
            visited.popitem()
            edges = edges[:-1]
        elif len(visited) < cutoff:
            if child == target:
                # Target has been found.
                yield list(edges) + [((child_source, child), child_key)]
            elif child not in visited:
                # Child node is not yet in the current path.
                visited[child] = None  # Mark as visited.
                # Append child node to the current path.
                edges.append(((child_source, child), child_key))
                # Put on the stack an iterator over all edges adjacent to the child node.
                stack.append(
                    (((u, v), k) for u, v, k in graph.edges(child, keys=True))
                )
        else:
            assert len(visited) == cutoff
            # TODO: Is this correct?
            count = (list(children) + [child]).count(target)
            for i in range(count):
                yield list(edges) + [((child_source, child), child_key)]
            stack.pop()
            visited.popitem()
            edges = edges[:-1]


def _get_conductivity_conditions(cmos_graph: nx.MultiGraph,
                                 inputs: Set,
                                 output_node) -> Dict[sympy.Symbol, boolalg.Boolean]:
    """
    For each input-output pair find the condition that there is a conductive path.

    For example an inverter with input A and output Y could be represented as a multi-graph:
    `VDD--[A]--Y--[A]--GND`

    The only output node would be 'Y', inputs would be the set {'A'}.
    The returned dictionary should then contain the condition that there is a conductive path from
    Y to VDD (A == 0) as well as the condition that there is a conductive path from Y to GND (A == 1).


    :param cmos_graph:
    :param vdd_node: Name of VDD supply node.
    :param gnd_node: Name of GND supply node.
    :param inputs: Set of all input pins including power pins.
    :param output_node:
    :return: sympy.Symbol
    """

    gate_input_nets = find_input_gates(cmos_graph)
    # Find input nets that also connect to a source or drain, i.e. to a transmission gate.
    transmission_input_pins = inputs - gate_input_nets
    logger.debug("Input pins to a transmission gate: {}".format(transmission_input_pins))
    if len(transmission_input_pins) < 2:
        logger.warning("`inputs` is expected to also contain VDD and GND.")

    def conductivity_condition(cmos_graph: nx.MultiGraph, source, target) -> boolalg.Boolean:
        """
        Find a boolean equation that evaluates to true iff there is a conductive path from `source` to `target`
        given the input signals.
        :param cmos_graph:
        :param source:
        :param target:
        :return: Boolean function. (sympy.Symbol)
        """
        # Get all simple paths from source to target.
        all_paths = list(_all_simple_paths_multigraph(cmos_graph, source, target))
        transistor_paths = [
            [(gate_net, channel_type) for (net1, net2), (gate_net, channel_type) in path] for path in all_paths
        ]
        # There is a conductive path if at least one of the paths is conductive -> Or
        # A path is conductive if all edges (transistors) along the path are conductive -> And
        f = boolalg.Or(
            *[boolalg.And(
                *(sympy.Symbol(gate) if channel_type == ChannelType.NMOS else ~sympy.Symbol(gate)
                  for gate, channel_type in path
                  )
            ) for path in transistor_paths]
        )
        # Try to simplify the boolean expression.
        # f = simplify_logic(f)
        logger.debug("Conductivity condition from '{}' to '{}': {}".format(source, target, f))
        return f

    def remove_nodes(graph: nx.MultiGraph, delete_nodes: Set) -> nx.MultiGraph:
        remaining_nodes = graph.nodes() - delete_nodes
        return graph.subgraph(remaining_nodes)

    # Calculate conductivity conditions from each input-pin
    # (i.e. power pins and inputs to transmission gates) to output.
    conductivity_conditions = {
        sympy.Symbol(i):
            conductivity_condition(
                # Remove all other nodes connected to transmission input pins.
                # This avoids for instance finding a path to GND leading over VDD.
                remove_nodes(cmos_graph, transmission_input_pins - {i}),
                output_node,
                i
            )
        for i in transmission_input_pins
    }

    return conductivity_conditions


def _high_impedance_condition(conductivity_conditions: Dict[sympy.Symbol, boolalg.Boolean]) -> boolalg.Boolean:
    pass


def _is_complementary(f1: boolalg.Boolean, f2: boolalg.Boolean) -> bool:
    """
    Check if two formulas are complementary.
    :param f1:
    :param f2:
    :return: Returns true iff `f1` and `f1` are complementary.
    """
    return bool_equals(f1, ~f2)


def complex_cmos_graph_to_formula(cmos_graph: nx.MultiGraph,
                                  output_nodes: Set,
                                  input_pins: Set) \
        -> Tuple[Dict[Any, Dict[sympy.Symbol, boolalg.Boolean]], set]:
    """
    Iteratively find formulas of intermediate nodes in complex gates which consist of multiple pull-up/pull-down networks.
    :param cmos_graph: The CMOS transistor network represented as a multi-graph.
    :param output_nodes: Set of output nodes. This corresponds to the output nets/pins of the cell.
    :param input_pins: Specify additional input pins that could otherwise not be deduced.
        This is mainly meant for power supply pins and inputs that connect not only to transistor gates but also source or drain.
    :return: (Dict[intermediate variable, expression for it], Set[input variables of the circuit])
    """

    # Sanity check: all output nodes must be in the graph.
    for n in output_nodes:
        assert n in cmos_graph.nodes(), "Output node is not in the graph: {}".format(n)

    # Sanity check: There must be at least two input pins.
    assert len(input_pins) >= 2, "`input_pins` must at least contain 2 nodes (GND, VDD). Found: {}".format(input_pins)
    # Make an independent copy that can be modified without causing side effects.
    input_pins = input_pins.copy()

    # Consider all nodes as input variables that are either connected to transistor
    # gates only or are specified by the caller.
    deduced_input_pins = find_input_gates(cmos_graph)
    logger.debug("Deduced input pins: {}".format(deduced_input_pins))
    inputs = deduced_input_pins | input_pins
    logger.debug("All input pins: {}".format(deduced_input_pins))

    # Set of nodes where there is no boolean formula known yet.
    unknown_nodes = {n for n in output_nodes}
    # Set of nodes for which a boolean expression is known.
    known_nodes = {i for i in inputs}
    # Boolean formulas for the outputs.
    output_formulas = dict()

    # Loop as long as there is a intermediary node with unknown value.
    while unknown_nodes:
        # Grab a node with unknown boolean expression.
        temp_output_node = unknown_nodes.pop()
        assert temp_output_node not in known_nodes
        # Follow the pull-up and pull-down paths to find the boolean conditions that
        # tell when there is a conductive path to nodes like VDD, GND (or inputs and internal nodes in case
        # of transmission gates).
        logger.debug("Find conductivity conditions for: {}".format(temp_output_node))
        conductivity_conditions = _get_conductivity_conditions(cmos_graph, input_pins, temp_output_node)

        output_formulas[temp_output_node] = conductivity_conditions
        known_nodes.add(temp_output_node)

        # Find all variables that occur in the conductivity conditions to later resolve them too.
        inputs_to_f = {a.name
                       for f in conductivity_conditions.values()
                       for a in f.atoms(sympy.Symbol)
                       }

        # Update the set of unknown variables.
        unknown_nodes = (unknown_nodes | inputs_to_f) - known_nodes

    return output_formulas, inputs


def test_complex_cmos_graph_to_formula():
    # Create CMOS network of a AND gate (NAND -> INV).
    g = nx.MultiGraph()
    g.add_edge('vdd', 'nand', ('a', ChannelType.PMOS))
    g.add_edge('vdd', 'nand', ('b', ChannelType.PMOS))
    g.add_edge('gnd', '1', ('a', ChannelType.NMOS))
    g.add_edge('1', 'nand', ('b', ChannelType.NMOS))

    g.add_edge('vdd', 'output', ('nand', ChannelType.PMOS))
    g.add_edge('gnd', 'output', ('nand', ChannelType.NMOS))

    conductivity_conditions, inputs = complex_cmos_graph_to_formula(g, output_nodes={'output'},
                                                                    input_pins={'vdd', 'gnd'})
    logger.debug(f"Conductivity conditions: \n{conductivity_conditions}")
    formulas = {sympy.Symbol(output): cc[sympy.Symbol('vdd')] for output, cc in conductivity_conditions.items()}

    # Convert from strings into sympy symbols.
    inputs = {sympy.Symbol(i) for i in inputs}
    logger.debug(f'formulas = {formulas}')
    logger.debug(f'inputs = {inputs}')

    def resolve_intermediate_variables(formulas: Dict[sympy.Symbol, sympy.Symbol],
                                       output: sympy.Symbol) -> boolalg.Boolean:
        f = formulas[output].copy()
        # TODO: detect loops
        while f.atoms(sympy.Symbol) - inputs:
            f = f.subs(formulas)
            f = simplify_logic(f)
        return f

    # Solve equation system for output.
    f = resolve_intermediate_variables(formulas, sympy.Symbol('output'))

    f = simplify_logic(f)

    # Verify that the deduced formula equals a NAND.
    a, b = sympy.symbols('a b')
    AND = (a & b)
    assert f.equals(AND), "Transformation of CMOS graph into formula failed."


def _resolve_intermediate_variables(formulas: Dict[sympy.Symbol, boolalg.Boolean],
                                    inputs: Set[sympy.Symbol],
                                    formula: boolalg.Boolean) -> boolalg.Boolean:
    """
    Simplify the formula `formula` by iterative substitution with the `formulas`.
    :param formulas: Formulas used for substitution as a dict.
    :param inputs: Define input atoms.
    :param formula: The start.
    :return: Return the formula with all substitutions applied.
    """

    # Set of symbols that will not be substituted.
    stop_atoms = set(inputs)

    while formula.atoms(sympy.Symbol) - stop_atoms:

        for a in formula.atoms(sympy.Symbol) - stop_atoms:
            stop_atoms.update(formula.atoms(sympy.Symbol))
            if a in formulas:
                formula = formula.subs({a: formulas[a]})
                formula = simplify_logic(formula)

    return formula


def test_resolve_intermediate_variables():
    a, b, c = sympy.symbols('a b c')
    formulas = {a: b ^ c, c: ~a}
    inputs = {b}
    r = _resolve_intermediate_variables(formulas, inputs, a)
    assert bool_equals(r, b ^ ~a)


def _formula_dependency_graph(formulas: Dict[sympy.Symbol, boolalg.Boolean]) -> nx.DiGraph:
    """
    Create a graph representing the dependencies of the variables/expressions.
    Used to detect feedback loops.
    :param formulas:
    :return:
    """
    dependency_graph = nx.DiGraph()
    for output, expression in formulas.items():
        if isinstance(expression, boolalg.Boolean):
            for atom in expression.atoms(sympy.Symbol):
                # Output depends on every variable (atom) in the expression.
                dependency_graph.add_edge(output, atom)
        elif isinstance(expression, bool):
            # Handle True and False constants.
            dependency_graph.add_edge(output, expression)
        else:
            assert False, "Type not supported: '{}'".format(type(expression))
    return dependency_graph


def analyze_circuit_graph(graph: nx.MultiGraph,
                          pins_of_interest: Set,
                          constant_input_pins: Dict[Any, bool] = None,
                          differential_inputs: Dict[str, str] = None,
                          user_input_nets: Set = None,
                          debug: bool = False
                          ) -> AbstractCircuit:
    """
    Analyze a CMOS transistor network and find boolean expressions for the output signals of the `pins_of_interest`.

    The transistor network is represented as a multi-graph, where nodes represent nets and edges represent transistors.
    The end points of an edge correspond to drain and source of the transistor. The label of the edge corresponds to the
    gate net of the transistor.

    :param graph:
    :type graph: nx.MultiGraph
    :param pins_of_interest:
    :param constant_input_pins: Define input pins that have a constant and known value such as `{'vdd': True, 'gnd': False}`.
    :type constant_input_pins: Dict[Any, bool]
    :param differential_inputs: Mapping from non-inverting inputs to their inverting inputs.
    :type differential_inputs: Dict[str, str]
    :param user_input_nets: A set of input pin names.
        Some inputs cannot automatically be found and must be provided by the user.
        Input nets that connect not only to transistor gates but also source or drains need to be specified manually.
        This can happen for cells containing transmission gates.
    :param debug: More debugging output.
    :return: (Dict['output pin', boolean formula], Dict['output pin', Memory])
    """
    #
    # import matplotlib.pyplot as plt
    # nx.draw_networkx(graph)
    # plt.draw()
    # plt.show()

    pins_of_interest = set(pins_of_interest)

    if constant_input_pins is None:
        constant_input_pins = dict()
    if differential_inputs is None:
        differential_inputs = dict()

    gate_nets = _get_gate_nets(graph)
    nets = set(graph.nodes())
    all_nets = gate_nets | nets

    # Sanity check for the inputs.
    non_present_pins = set()
    for p in pins_of_interest:
        if p not in all_nets:
            logger.warning(f"Net '{p}' is not connected to any transistor.")
            non_present_pins.add(p)

    pins_of_interest = pins_of_interest - non_present_pins

    logger.info("VDD nets: {}".format(", ".join([p for p, v in constant_input_pins.items() if v])))
    logger.info("GND nets: {}".format(", ".join([p for p, v in constant_input_pins.items() if not v])))

    if user_input_nets is None:
        user_input_nets = set()
    else:
        user_input_nets = set(user_input_nets)

    logger.info("User supplied input nets: {}".format(user_input_nets))
    deduced_input_nets = find_input_gates(graph)
    logger.info("Additional detected input nets: {}".format(deduced_input_nets))
    input_nets = user_input_nets | deduced_input_nets | constant_input_pins.keys()
    output_nodes = pins_of_interest - input_nets
    logger.info("Detected input nets: {}".format(input_nets))
    logger.info("Detected output nets: {}".format(output_nodes))

    # Find conductivity conditions for all pull-up/pull-down paths.
    conductivity_conditions, inputs = complex_cmos_graph_to_formula(graph,
                                                                    input_pins=input_nets,
                                                                    output_nodes=output_nodes,
                                                                    )
    # print('conductivity_conditions = ', conductivity_conditions)

    # Pretty-print the conductivity conditions.
    if debug:
        print()
        print('Conductivity conditions')
        for pin_a, conditions in conductivity_conditions.items():
            print(' ', pin_a)
            for pin_b, condition in conditions.items():
                print('  |-', pin_b, 'when', condition)
        print()

    # Substitute inverting inputs by the inverse of the non-inverting input.
    differential_substitution = {
        sympy.Symbol(inv): ~sympy.Symbol(noninv)
        for noninv, inv in differential_inputs.items()
    }

    # Add known values for VDD, GND
    constants = {sympy.Symbol(k): v for k, v in constant_input_pins.items()}
    substitutions = dict()
    substitutions.update(differential_substitution)
    substitutions.update(constants)
    output_symbols = {sympy.Symbol(n) for n in output_nodes}

    # Convert keys into symbols.
    conductivity_conditions = {sympy.Symbol(k): v for k, v in conductivity_conditions.items()}

    # Derive logic formulas from conductivity condition.
    logger.debug("Derive logic formulas from conductivity condition.")
    formulas_high = dict()
    formulas_low = dict()
    for output, cc in conductivity_conditions.items():
        logger.debug(f"Derive logic formulas from conductivity condition for '{output}'.")
        or_terms_high = []
        or_terms_low = []
        for input_pin, condition in cc.items():
            assert isinstance(input_pin, sympy.Symbol)
            assert isinstance(condition, boolalg.Boolean)
            # Simplify by inserting constants.
            condition = simplify_logic(condition.subs(substitutions))
            input_pin = simplify_logic(input_pin.subs(substitutions))
            # if `condition` then output is `connected` to `input_pin`.
            connected_to_high = simplify_logic(condition & input_pin)  # Is there a path to HIGH?
            connected_to_low = simplify_logic(condition & ~input_pin)  # Is there a path to LOW?
            or_terms_high.append(connected_to_high)
            or_terms_low.append(connected_to_low)

        # Logic OR of all elements.
        formulas_high[output] = simplify_logic(sympy.Or(*or_terms_high))
        formulas_low[output] = simplify_logic(sympy.Or(*or_terms_low))

    logger.debug('formulas_high = {}'.format(formulas_high))
    logger.debug('formulas_low = {}'.format(formulas_low))

    # Convert from strings into sympy symbols.
    inputs = {sympy.Symbol(i) for i in inputs} - set(constants.keys())
    logger.debug('inputs = {}'.format(inputs))

    # Find formulas for nets that are complementary and can never be in a high-impedance nor short-circuit state.
    logger.debug("Find formulas for nets that are complementary and "
                 "can never be in a high-impedance nor short-circuit state.")
    complementary_formulas = {k: formulas_high[k]
                              for k in formulas_high.keys()
                              if _is_complementary(formulas_high[k], formulas_low[k])}

    # Use the complementary nets to simplify the other formulas.
    formulas_high = {n: f.subs(complementary_formulas) for n, f in formulas_high.items()}
    formulas_low = {n: f.subs(complementary_formulas) for n, f in formulas_low.items()}

    # Find high-Z conditions.
    logger.debug("Find high-Z conditions.")
    high_impedance_conditions = {k: simplify_logic(~formulas_high[k] & ~formulas_low[k])
                                 for k in formulas_high.keys()}

    # Find short-circuit conditions.
    logger.debug("Find short-circuit conditions.")
    short_circuit_conditions = {k: simplify_logic(formulas_high[k] & formulas_low[k])
                                for k in formulas_high.keys()}

    logger.debug("Complementary nets:")
    for net, f in complementary_formulas.items():
        logger.debug(f' {net}: {f}')

    logger.debug(f"High impedance conditions ({len(high_impedance_conditions)}):")
    for net, condition in high_impedance_conditions.items():
        msg = f"  '{net}' is high-impedance when '{condition}'"
        if satisfiable(condition):
            logger.warning(msg)
        else:
            logger.debug(msg)

    logger.debug(f"Short circuit conditions (len(short_circuit_conditions)):")
    for net, condition in short_circuit_conditions.items():
        msg = f" short circuit on '{net}' when '{condition}'"
        if satisfiable(condition):
            logger.warning(msg)
        else:
            logger.debug(msg)

    # Create dependency graph to detect feedback loops.
    dependency_graph_high = _formula_dependency_graph(formulas_high)
    dependency_graph_low = _formula_dependency_graph(formulas_low)
    # dependency_graph = nx.DiGraph(itertools.chain(dependency_graph_high.edges, dependency_graph_low.edges))
    dependency_graph = dependency_graph_high

    # import matplotlib.pyplot as plt
    # nx.draw_networkx(dependency_graph)
    # plt.draw()
    # plt.show()

    # Check for cycles.
    cycles = list(nx.simple_cycles(dependency_graph))
    logger.info("Number of feed-back loops: {}".format(len(cycles)))

    # # Print feedback loops.
    # for cycle in cycles:
    #     print()
    #     print("cycle: {}".format(cycle))
    #     for el in cycle:
    #         print(' --> {} = {}'.format(el, formulas_high[el]))
    #     print()

    # Collect all nets that belong to a memory cycle.
    nets_of_feedback_loops = {n for c in cycles for n in c}
    logger.debug(f"Nets of feed-back loops: {nets_of_feedback_loops}")

    def derive_memory(memory_output_net: sympy.Symbol,
                      inputs: Set[sympy.Symbol],
                      formulas: Dict[sympy.Symbol, boolalg.Boolean]) -> Memory:
        """
        Detect a memory loop given the output net of the memory.
        :param memory_output_net:
        :param inputs: Input atoms (the input nets to the loop).
        :param formulas:
        :return:
        """
        logger.debug("Derive memory from output net: {}".format(memory_output_net))
        # print("Derive memory from output net: {}".format(memory_output_net))
        memory_output_resolved = _resolve_intermediate_variables(formulas, inputs, memory_output_net)
        # print(formulas)
        # print('memory_output_resolved =', memory_output_resolved)

        logger.debug("Memory output net {}  = {}".format(memory_output_net, memory_output_resolved))
        d, dp, dn = boolean_derivatives(memory_output_resolved, memory_output_net)
        write_condition = ~dp
        logger.debug(f"write_condition = {write_condition}")
        # print(f"write_condition = {write_condition}")
        oscillation_condition = dn

        if debug:
            # Find expressions for the memory output when the write condition is met.
            # Find all variable assignments such that the write condition is met.
            write_condition_models = list(satisfiable(write_condition, all_models=True))
            logger.debug(f"write_condition_models = {write_condition_models}")
            if write_condition_models == [False]:
                logger.warning("Detected a memory loop that is not possible to write to.")

            # Find what data the memory will store for each write condition.
            mem_data = []
            for model in write_condition_models:
                # Now find the memory output once the write condition is met.
                data = simplify_logic(memory_output_resolved.subs(model))
                mem_data.append((model, data))

            if debug:
                # Pretty print write conditions and corresponding write data.
                if len(mem_data) > 0:
                    print()
                    print("Write data for all write conditions:")
                    model, _ = mem_data[0]
                    vars = list(model.keys())
                    print("\t", "\t".join((str(v) for v in vars)), "\t:", f"data ({memory_output_net})")
                    for model, data in mem_data:
                        print("\t", "\t".join((str(model[var]) for var in vars)), "\t:", data)
                    print()

        write_data = simplify_with_assumption(write_condition, memory_output_resolved)
        logger.debug(f"write_data = {write_data}")

        if not sympy.satisfiable(~write_condition):
            logger.warning("This is not a true memory, it will never store anything.")

        return Memory(data=write_data, write_condition=write_condition, oscillation_condition=oscillation_condition)

    # Derive memory for all cycles with each possible node as memory output.
    memory_output_nets = set()
    memory_by_output_net = dict()  # Cache memory loops for later usage.
    nets_of_memory_loops = set()
    for loop_id, cycle in enumerate(cycles):
        # print('cycle = {}'.format(cycle))
        nodes_in_cycle = set(cycle)
        _inputs = inputs | nets_of_feedback_loops - nodes_in_cycle
        for node in cycle:
            memory = derive_memory(node, _inputs, formulas_high)
            memory.loop_id = loop_id

            # print(memory)
            # print()

            if sympy.satisfiable(~memory.write_condition):
                # Store condition can be met -> this is a potential memory.
                memory_output_nets.add(node)
                memory_by_output_net[node] = memory
                nets_of_memory_loops.update(cycle)

    logger.debug("Memory output nets: {}".format(memory_output_nets))

    # Solve equation system for output.
    # TODO: stop resolving at memory elements.
    def solve_output_nets() -> Tuple[Dict[boolalg.BooleanAtom, boolalg.Boolean], Dict[boolalg.BooleanAtom, Memory]]:
        output_formulas = dict()
        latches = dict()

        wavefront = {sympy.Symbol(o) for o in output_nodes}

        while wavefront:

            out = wavefront.pop()
            assert out not in output_formulas
            assert out not in latches

            new_wavefront = set()

            if out not in nets_of_memory_loops:
                logger.debug(f"Find formula for {out}.")

                assert isinstance(out, boolalg.Boolean)
                formula = _resolve_intermediate_variables(formulas_high, inputs | nets_of_memory_loops, out)
                output_formulas[out] = formula

                # Find new wavefront.
                atoms = formula.atoms(sympy.Symbol)
                new_wavefront.update(atoms)

                # Find inputs into formula that come from a memory.
                new_atoms = atoms - output_formulas.keys() - latches.keys() - inputs
                unknown_memory_nets = new_atoms & nets_of_memory_loops
                logger.debug(f"Unknown inputs into {out} from memory: {unknown_memory_nets}")
            else:
                # Resolve memory.
                assert out in nets_of_memory_loops

                # Output net of the memory loop.
                memory_net = out
                memory = memory_by_output_net[memory_net]

                latches[memory_net] = memory

                # Find new wavefront.
                atoms = memory.write_condition.atoms(sympy.Symbol) | memory.data.atoms(sympy.Symbol)
                new_wavefront.update(atoms)

            wavefront.update(new_wavefront - output_formulas.keys() - latches.keys() - inputs)

        return output_formulas, latches

    output_formulas, latches = solve_output_nets()

    # Simplify formulas in memories.
    for m in latches.values():
        m.data = simplify_logic(m.data.subs(output_formulas))
        m.write_condition = simplify_logic(m.write_condition.subs(output_formulas))
        m.oscillation_condition = simplify_logic(m.oscillation_condition.subs(output_formulas))

    logger.debug("Output functions ({}):".format(len(output_formulas)))
    output_combinatorial = dict()
    for net, formula in output_formulas.items():
        z = high_impedance_conditions[net]

        # Get the output function assuming that the output is not high-impedance.
        function = simplify_with_assumption(~z, formula)

        output_combinatorial[net] = CombinationalOutput(function=function, high_impedance=z)
        logger.debug(f"{net} = {function}, Z: {z}")

    logger.debug("Latches ({}):".format(len(latches)))

    for output_net, latch in latches.items():
        logger.info(f"{output_net} = {latch}")

    if not latches:
        logger.info("No latches found.")

    if len(cycles) > 0 and len(latches) == 0:
        logger.warning("Found feedback loops but no latches were derived.")
    if len(latches):
        logger.info(f"Number of latches: {len(latches)}")

    return AbstractCircuit(output_symbols, output_combinatorial, latches)


class NetlistGen:
    """
    Generate test netlists.
    """

    def __init__(self):
        self.net_counter = itertools.count()
        self.vdd = "vdd"
        self.gnd = "gnd"

    def new_net(self, prefix=""):
        return "{}{}".format(prefix, next(self.net_counter))

    def inv(self, a, out):
        # Create transistors of a INV.
        return [
            (self.vdd, out, (a, ChannelType.PMOS)),
            (self.gnd, out, (a, ChannelType.NMOS)),
        ]

    def nor2(self, a, b, out):
        # Create transistors of a NOR.
        i = self.new_net("i")
        return [
            (self.vdd, i, (a, ChannelType.PMOS)),
            (i, out, (b, ChannelType.PMOS)),
            (out, self.gnd, (a, ChannelType.NMOS)),
            (out, self.gnd, (b, ChannelType.NMOS)),
        ]

    def nor3(self, a, b, c, out):
        # Create transistors of a NOR.
        i1 = self.new_net("i1")
        i2 = self.new_net("i2")
        return [
            (self.vdd, i1, (a, ChannelType.PMOS)),
            (i1, i2, (b, ChannelType.PMOS)),
            (i2, out, (c, ChannelType.PMOS)),
            (out, self.gnd, (a, ChannelType.NMOS)),
            (out, self.gnd, (b, ChannelType.NMOS)),
            (out, self.gnd, (c, ChannelType.NMOS)),
        ]

    def nand2(self, a, b, out):
        # Create transistors of a NAND.
        i = self.new_net("i")
        return [
            (self.vdd, out, (a, ChannelType.PMOS)),
            (self.vdd, out, (b, ChannelType.PMOS)),
            (out, i, (a, ChannelType.NMOS)),
            (i, self.gnd, (b, ChannelType.NMOS)),
        ]

    def nand3(self, a, b, c, out):
        # Create transistors of a NAND.
        i1 = self.new_net("i1")
        i2 = self.new_net("i2")
        return [
            (self.vdd, out, (a, ChannelType.PMOS)),
            (self.vdd, out, (b, ChannelType.PMOS)),
            (self.vdd, out, (c, ChannelType.PMOS)),
            (out, i1, (a, ChannelType.NMOS)),
            (i1, i2, (c, ChannelType.NMOS)),
            (i2, self.gnd, (b, ChannelType.NMOS)),
        ]

    def and2(self, a, b, out):
        nand_out = self.new_net(prefix="nand_out")
        return self.nand2(a, b, nand_out) + self.inv(nand_out, out)

    def and3(self, a, b, c, out):
        nand_out = self.new_net(prefix="nand_out")
        return self.nand3(a, b, c, nand_out) + self.inv(nand_out, out)

    def or2(self, a, b, out):
        nor_out = self.new_net(prefix="nor_out")
        return self.nor2(a, b, nor_out) + self.inv(nor_out, out)

    def or3(self, a, b, c, out):
        nor_out = self.new_net(prefix="nor_out")
        return self.nor3(a, b, c, nor_out) + self.inv(nor_out, out)

    def mux2(self, in0, in1, sel, out):
        sel0 = self.new_net(prefix='sel0')
        sel1 = sel
        in0_masked = self.new_net('in0_masked')
        in1_masked = self.new_net('in1_masked')

        return self.inv(sel1, sel0) + self.nand2(in0, sel0, in0_masked) + self.nand2(in1, sel1, in1_masked) + \
               self.nand2(in0_masked, in1_masked, out)

    def latch(self, clk, d_in, d_out):
        a_2_6 = self.new_net("a_2_6_")
        a_18_74 = self.new_net("a_18_74_")
        a_23_6 = self.new_net("a_23_6_")
        a_35_84 = self.new_net("a_35_84_")
        a_18_6 = self.new_net("a_18_6_")
        a_35_6 = self.new_net("a_35_6_")

        edges = [
            # Command for converting a raw SPICE transistor netlist into this format:
            # cat raw_netlist.sp | grep -v + | awk '{ print "(\"" $2 "\", \"" $4 "\", (\"" $3  "\", " $6 ")),"}' \
            # | sed 's/pmos/ChannelType.PMOS/g' | sed 's/nmos/ChannelType.NMOS/g'

            (self.vdd, a_2_6, (clk, ChannelType.PMOS)),
            (a_18_74, self.vdd, (d_in, ChannelType.PMOS)),
            (a_23_6, a_18_74, (a_2_6, ChannelType.PMOS)),
            (a_35_84, a_23_6, (clk, ChannelType.PMOS)),
            (self.vdd, a_35_84, (d_out, ChannelType.PMOS)),
            (self.gnd, a_2_6, (clk, ChannelType.NMOS)),
            (d_out, self.vdd, (a_23_6, ChannelType.PMOS)),
            (a_18_6, self.gnd, (d_in, ChannelType.NMOS)),
            (a_23_6, a_18_6, (clk, ChannelType.NMOS)),
            (a_35_6, a_23_6, (a_2_6, ChannelType.NMOS)),
            (self.gnd, a_35_6, (d_out, ChannelType.NMOS)),
            (d_out, self.gnd, (a_23_6, ChannelType.NMOS)),
        ]

        return edges


def test_analyze_circuit_graph():
    # Create CMOS network of a AND gate (NAND -> INV).

    gen = NetlistGen()
    edges = gen.and2('a', 'b', 'output')
    g = nx.MultiGraph()
    for e in edges:
        g.add_edge(*e)

    pins_of_interest = {'output'}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest, constant_input_pins=known_pins)

    # Verify that the deduced boolean function is equal to the AND function.
    a, b = sympy.symbols('a, b')
    assert bool_equals(abstract.outputs[sympy.Symbol('output')].function, a & b)


def test_analyze_circuit_graph_transmission_gate_xor():
    g = nx.MultiGraph()
    # Represent an XOR gate with transmission-gates in its graph form.
    # Edges correspond to transistors, nodes to nets. The 'key' of the edge
    # Contains a Tuple ('net name of the gate', 'channel type').

    # Inverter A
    g.add_edge('vdd', 'a_not', ('a', ChannelType.PMOS))
    g.add_edge('gnd', 'a_not', ('a', ChannelType.NMOS))
    # Inverter B
    g.add_edge('vdd', 'b_not', ('b', ChannelType.PMOS))
    g.add_edge('gnd', 'b_not', ('b', ChannelType.NMOS))
    # Transmission gates
    g.add_edge('a_not', 'c', ('b', ChannelType.NMOS))
    g.add_edge('a_not', 'c', ('b_not', ChannelType.PMOS))
    g.add_edge('a', 'c', ('b_not', ChannelType.NMOS))
    g.add_edge('a', 'c', ('b', ChannelType.PMOS))

    # Pin 'a' must be explicitly given as an input pin.
    # It connects not only to gates but also to source/drains of transistors
    # and therefore cannot be deduced to be an input pin.
    pins_of_interest = {'a', 'b', 'c'}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest,
                                     constant_input_pins=known_pins,
                                     user_input_nets={'a', 'b'})

    # Verify that the deduced boolean function is equal to the XOR function.
    a, b = sympy.symbols('a, b')
    assert bool_equals(abstract.outputs[sympy.Symbol('c')].function, a ^ b)
    assert not abstract.latches


def test_analyze_circuit_graph_mux2():
    gen = NetlistGen()

    edges = []

    d0 = 'a'
    d1 = 'b'
    out = 'y'
    sel = 's'

    edges += gen.mux2(d0, d1, sel, out)

    g = nx.MultiGraph()
    for e in edges:
        g.add_edge(*e)

    pins_of_interest = {d0, d1, out, sel}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest, constant_input_pins=known_pins)
    # Verify that the deduced boolean function is equal to the MUX function.

    a, b, s = sympy.symbols('a, b, s')
    assert bool_equals(abstract.outputs[sympy.Symbol('y')].function, (a & ~s) | (b & s))
    assert not abstract.latches


def test_analyze_circuit_graph_latch():
    g = nx.MultiGraph()
    # Network of a LATCH.
    gen = NetlistGen()
    edges = gen.latch('CLK', 'D', 'Q')
    for e in edges:
        g.add_edge(*e)

    pins_of_interest = {'CLK', 'D', 'Q'}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest, constant_input_pins=known_pins)

    assert len(abstract.latches) == 1


def test_analyze_circuit_graph_set_reset_nand():
    # Network of a set-reset nand feedback loop.
    r"""
     S -----|\
            |&O--+--- Y1
        +---|/   |
        |        |
        | +------+
        | |
        +-|------+
          |      |
          +-|\   |
            |&O--+--- Y2
     R -----|/
    """

    gen = NetlistGen()

    g = nx.MultiGraph()
    edges = gen.nand2("S", "Y2", "Y1") + gen.nand2("R", "Y1", "Y2")
    for e in edges:
        g.add_edge(*e)

    pins_of_interest = {'S', 'R', 'Y1', 'Y2'}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest, constant_input_pins=known_pins)
    print(abstract.outputs)
    print(abstract.latches)

    assert len(abstract.latches) == 1
    # TODO


def test_analyze_circuit_graph_2w11():
    # Gate with hysteresis.
    # Output goes HIGH when two inputs are HIGH.
    # Output goes LOW when both inputs are LOW. Otherwise output is hold.
    g = nx.MultiGraph()
    # Network of a LATCH.
    gen = NetlistGen()
    edges = gen.nand2("S", "Y2", "Y1") + gen.nand2("R", "Y1", "Y2")
    edges += gen.nand2('ip1', 'ip2', 'S')
    edges += gen.or2('ip1', 'ip2', 'R')
    for e in edges:
        g.add_edge(*e)

    pins_of_interest = {'ip1', 'ip2', 'Y1'}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest, constant_input_pins=known_pins)
    print(abstract.latches)
    assert len(abstract.latches) == 1

    ip1, ip2, y1 = sympy.symbols('ip1 ip2 Y1')
    latch = abstract.latches[y1]
    assert isinstance(latch, Memory)
    assert bool_equals(latch.write_condition, (ip1 & ip2) | (~ip1 & ~ip2))


def test_analyze_circuit_graph_dff_pos():
    gen = NetlistGen()

    edges = []

    d = 'D'
    q = 'Q'
    clk = 'CLK'
    clk_inv = gen.new_net(prefix='clk_inv')
    d_i = gen.new_net(prefix='d_i')

    edges += gen.inv(clk, clk_inv)
    edges += gen.latch(clk, d, d_i)
    edges += gen.latch(clk_inv, d_i, q)

    g = nx.MultiGraph()
    for e in edges:
        g.add_edge(*e)

    pins_of_interest = {clk, d, q}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest, constant_input_pins=known_pins)

    assert len(abstract.latches) == 2


def test_analyze_circuit_graph_dff_pos_sync_reset():
    gen = NetlistGen()

    edges = []

    d = 'D'
    q = 'Q'
    r = 'R'
    clk = 'CLK'
    clk_inv = gen.new_net(prefix='clk_inv')
    d_i = gen.new_net(prefix='d_i')
    d_rst = gen.new_net(prefix='d_rst')

    edges += gen.and2(d, r, d_rst)
    edges += gen.inv(clk, clk_inv)
    edges += gen.latch(clk_inv, d_rst, d_i)
    edges += gen.latch(clk, d_i, q)

    g = nx.MultiGraph()
    for e in edges:
        g.add_edge(*e)

    pins_of_interest = {clk, d, q}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest, constant_input_pins=known_pins)

    assert len(abstract.latches) == 2


def test_analyze_circuit_graph_dff_pos_scan():
    gen = NetlistGen()

    edges = []

    d = 'D'
    q = 'Q'
    clk = 'CLK'
    scan_enable = 'ScanEnable'
    scan_in = 'ScanIn'
    scan_mux_out = 'ScanMux_DO'
    clk_inv = gen.new_net(prefix='clk_inv')
    d_i = gen.new_net(prefix='d_i')

    edges += gen.inv(clk, clk_inv)

    edges += gen.mux2(d, scan_in, scan_enable, scan_mux_out)

    edges += gen.latch(clk, scan_mux_out, d_i)
    edges += gen.latch(clk_inv, d_i, q)

    g = nx.MultiGraph()
    for e in edges:
        g.add_edge(*e)

    pins_of_interest = {clk, d, q}
    known_pins = {'vdd': True, 'gnd': False}
    abstract = analyze_circuit_graph(g, pins_of_interest=pins_of_interest, constant_input_pins=known_pins)
    assert len(abstract.latches) == 2
